import string
from unittest import mock

import flask
import flask_jwt_extended
import jwt
import ldap
import pytest

from keg_auth.libs import authenticators as auth, get_domain_from_email
from keg_auth.tests.utils import oauth_profile
from keg_auth_ta.model.entities import User, UserNoEmail


class TestKegAuthenticator:
    def test_user_not_found(self):
        with pytest.raises(auth.UserNotFound):
            authenticator = auth.KegAuthenticator(app=flask.current_app)
            authenticator.verify_user(login_id='nobodybythisnamehere')

    def test_user_not_active(self):
        user = User.fake(is_enabled=False)
        with pytest.raises(auth.UserInactive) as e_info:
            authenticator = auth.KegAuthenticator(app=flask.current_app)
            authenticator.verify_user(login_id=user.email)
        assert e_info.value.user is user

    def test_user_bad_password(self):
        user = User.fake()
        with pytest.raises(auth.UserInvalidAuth) as e_info:
            authenticator = auth.KegAuthenticator(app=flask.current_app)
            authenticator.verify_user(login_id=user.email, password='cannotpossiblybethis')
        assert e_info.value.user is user

    def test_user_verified(self):
        user = User.fake()
        authenticator = auth.KegAuthenticator(app=flask.current_app)
        found_user = authenticator.verify_user(login_id=user.email, password=user._plaintext_pass)
        assert user is found_user

    def test_user_excluded(self):
        user = User.fake()
        authenticator = auth.KegAuthenticator(app=flask.current_app)
        authenticator.domain_exclusions = [get_domain_from_email(user.email)]
        with pytest.raises(auth.UserNotFound):
            authenticator.verify_user(login_id=user.email, password=user._plaintext_pass)

    def test_unverified_user(self):
        user = User.fake()
        user.is_verified = False
        authenticator = auth.KegAuthenticator(app=flask.current_app)
        with pytest.raises(auth.UserInactive) as e_info:
            authenticator.verify_user(login_id=user.email, password=user._plaintext_pass)
        assert e_info.value.user is user

        found_user = authenticator.verify_user(login_id=user.email, password=user._plaintext_pass,
                                               allow_unverified=True)
        assert user is found_user

    def test_user_case_insensitive(self):
        from keg import db
        from sqlalchemy import text

        user = User.fake(email='abc@foo.bar')

        # Downstream in the process of creating a user, the email will be set to all lowercase
        # So we need to manually set it to capital letters to test the fix
        with db.db.engine.connect() as connection:
            connection.execute(
                text("UPDATE users SET email = 'ABC@Foo.Bar' WHERE email = 'abc@foo.bar'")
            )
        authenticator = auth.KegAuthenticator(app=flask.current_app)
        found_user = authenticator.verify_user(login_id='aBc@Foo.Bar',
                                               password=user._plaintext_pass)
        assert found_user is not None
        assert found_user.id == user.id

    @mock.patch.dict(
        'flask.current_app.config',
        {
            'KEGAUTH_OAUTH_PROFILES': [
                oauth_profile(domain_filter='bar.baz'),
                oauth_profile(domain_filter=('foo.co', 'foo.mo')),
            ]
        }
    )
    def test_loads_oauth_exclusions(self):
        authenticator = auth.KegAuthenticator(app=flask.current_app)
        assert authenticator.domain_exclusions == ['bar.baz', 'foo.co', 'foo.mo']

    def test_domain_excluded(self):
        authenticator = auth.KegAuthenticator(app=flask.current_app)
        assert not authenticator.is_domain_excluded('foo@bar.baz')
        authenticator.domain_exclusions = ['bar.baz', 'foo.co', 'foo.mo']
        assert authenticator.is_domain_excluded('foo@bar.baz')
        assert not authenticator.is_domain_excluded('foo@bar.co')


class TestOAuthAuthenticator:
    @mock.patch.dict('flask.current_app.config', {'KEGAUTH_OAUTH_PROFILES': [oauth_profile()]})
    def test_profiles_loaded(self):
        auth.OAuthAuthenticator(app=flask.current_app)
        assert flask.current_app.auth_manager.oauth.create_client('google')
        assert not flask.current_app.auth_manager.oauth.create_client('twitter')

    @mock.patch.dict(
        'flask.current_app.config',
        {'KEGAUTH_OAUTH_PROFILES': [oauth_profile(oauth_client_kwargs={'name': 'foo'})]}
    )
    @mock.patch('authlib.integrations.flask_client.OAuth', autospec=True, spec_set=True)
    def test_kwargs_passed_to_oauth_client(self, m_oauth):
        auth.OAuthAuthenticator(app=flask.current_app)
        m_oauth.return_value.register.assert_called_once_with(name='foo')

    @mock.patch.dict('flask.current_app.config', {'KEGAUTH_OAUTH_PROFILES': [oauth_profile()]})
    def test_user_not_found(self):
        with pytest.raises(auth.UserNotFound):
            authenticator = auth.OAuthAuthenticator(app=flask.current_app)
            authenticator.verify_user(
                profile_name='google', login_id='nobodybythisnamehere@mycompany.biz'
            )

    @mock.patch.dict('flask.current_app.config', {'KEGAUTH_OAUTH_PROFILES': [oauth_profile()]})
    def test_user_not_active(self):
        user = User.fake(is_enabled=False, email='usernotfound@mycompany.biz')
        with pytest.raises(auth.UserInactive) as e_info:
            authenticator = auth.OAuthAuthenticator(app=flask.current_app)
            authenticator.verify_user(profile_name='google', login_id=user.email)
        assert e_info.value.user is user

    @mock.patch.dict('flask.current_app.config', {'KEGAUTH_OAUTH_PROFILES': [oauth_profile()]})
    def test_user_verified(self):
        user = User.fake(email='userverified@mycompany.biz')
        authenticator = auth.OAuthAuthenticator(app=flask.current_app)
        found_user = authenticator.verify_user(profile_name='google', login_id=user.email)
        assert user is found_user

    @mock.patch.dict('flask.current_app.config', {'KEGAUTH_OAUTH_PROFILES': [oauth_profile()]})
    def test_user_unverified(self):
        user = User.fake(email='userunverified@mycompany.biz')
        user.is_verified = False
        authenticator = auth.OAuthAuthenticator(app=flask.current_app)
        found_user = authenticator.verify_user(profile_name='google', login_id=user.email)
        assert user is found_user
        assert user.is_verified

    @mock.patch.dict('flask.current_app.config', {'KEGAUTH_OAUTH_PROFILES': [oauth_profile()]})
    def test_domain_exclusion(self):
        user = User.fake(email='userverified@someothercompany.co')
        authenticator = auth.OAuthAuthenticator(app=flask.current_app)
        with pytest.raises(auth.UserNotFound):
            authenticator.verify_user(profile_name='google', login_id=user.email)

    def test_bad_profile(self):
        user = User.fake()
        authenticator = auth.OAuthAuthenticator(app=flask.current_app)
        with pytest.raises(Exception, match='.*foo is not configured'):
            authenticator.verify_user(profile_name='foo', login_id=user.email)


class TestLdapAuthenticator:
    def setup_method(self):
        flask.current_app.config['KEGAUTH_LDAP_SERVER_URL'] = 'abc123'
        flask.current_app.config['KEGAUTH_LDAP_DN_FORMAT'] = '{}'

    @mock.patch('ldap.initialize', autospec=True, spec_set=True)
    def test_user_not_found(self, mocked_ldap):
        mocked_ldap.return_value.simple_bind_s.return_value = (ldap.RES_BIND, )

        authenticator = auth.LdapAuthenticator(app=flask.current_app)
        success = authenticator.verify_user(login_id='nobodybythisnamehere', password='foo')
        assert mocked_ldap.call_count
        assert success
        assert User.get_by(username='nobodybythisnamehere')

    @mock.patch('ldap.initialize', autospec=True, spec_set=True)
    def test_user_not_active(self, mocked_ldap):
        # internal flag should have no effect on LDAP auth
        mocked_ldap.return_value.simple_bind_s.return_value = (ldap.RES_BIND, )
        user = User.fake(is_enabled=False)
        authenticator = auth.LdapAuthenticator(app=flask.current_app)
        assert authenticator.verify_user(login_id=user.email)

    @mock.patch('ldap.initialize', autospec=True, spec_set=True)
    def test_no_server_url_set(self, mocked_ldap):
        del flask.current_app.config['KEGAUTH_LDAP_SERVER_URL']

        user = User.fake()
        authenticator = auth.LdapAuthenticator(app=flask.current_app)
        with pytest.raises(Exception) as e_info:
            authenticator.verify_password(user, None)
        assert 'KEGAUTH_LDAP_SERVER_URL' in str(e_info.value)

    @mock.patch('ldap.initialize', autospec=True, spec_set=True)
    def test_no_dn_format_set(self, mocked_ldap):
        del flask.current_app.config['KEGAUTH_LDAP_DN_FORMAT']

        user = User.fake()
        authenticator = auth.LdapAuthenticator(app=flask.current_app)
        with pytest.raises(Exception) as e_info:
            authenticator.verify_password(user, None)
        assert 'KEGAUTH_LDAP_DN_FORMAT' in str(e_info.value)

    @mock.patch('ldap.initialize', autospec=True, spec_set=True)
    def test_unsuccessful_authentication(self, mocked_ldap):
        mocked_ldap.return_value.simple_bind_s.side_effect = ldap.INVALID_CREDENTIALS()

        user = User.fake()
        authenticator = auth.LdapAuthenticator(app=flask.current_app)
        success = authenticator.verify_password(user, 'foo')

        assert mocked_ldap.call_count
        assert success is False

    @mock.patch('ldap.initialize', autospec=True, spec_set=True)
    def test_invalid_dn_syntax(self, mocked_ldap):
        mocked_ldap.return_value.simple_bind_s.side_effect = ldap.INVALID_DN_SYNTAX()

        user = User.fake()
        authenticator = auth.LdapAuthenticator(app=flask.current_app)
        success = authenticator.verify_password(user, 'foo')

        assert mocked_ldap.call_count
        assert success is False

    @mock.patch('ldap.initialize', autospec=True, spec_set=True)
    def test_unsuccessful_authentication_wrong_result(self, mocked_ldap):
        mocked_ldap.return_value.simple_bind_s.return_value = (0, )

        user = User.fake()
        authenticator = auth.LdapAuthenticator(app=flask.current_app)
        success = authenticator.verify_password(user, 'foo')

        assert mocked_ldap.call_count
        assert success is False

    @mock.patch('ldap.initialize', autospec=True, spec_set=True)
    def test_successful_authentication(self, mocked_ldap):
        mocked_ldap.return_value.simple_bind_s.return_value = (ldap.RES_BIND, )

        user = User.fake()
        authenticator = auth.LdapAuthenticator(app=flask.current_app)
        success = authenticator.verify_password(user, 'foo')

        assert mocked_ldap.call_count
        assert success is True

    @mock.patch('ldap.initialize', autospec=True, spec_set=True)
    def test_successful_authentication_multiple_server_urls(self, mocked_ldap):
        flask.current_app.config['KEGAUTH_LDAP_SERVER_URL'] = ['abc123', 'def456', 'ghi789']
        mocked_ldap.return_value.simple_bind_s.side_effect = (
            (0,),
            (0,),
            (ldap.RES_BIND,)
        )

        user = User.fake()
        authenticator = auth.LdapAuthenticator(app=flask.current_app)
        success = authenticator.verify_password(user, 'foo')

        assert mocked_ldap.call_args_list == [
            mock.call('abc123'),
            mock.call('def456'),
            mock.call('ghi789'),
        ]
        assert success is True

    @mock.patch('ldap.initialize', autospec=True, spec_set=True)
    def test_debug_override(self, mocked_ldap):
        flask.current_app.config['KEGAUTH_LDAP_TEST_MODE'] = True

        user = User.fake()
        authenticator = auth.LdapAuthenticator(app=flask.current_app)
        success = authenticator.verify_password(user, 'foo')

        assert not mocked_ldap.call_count
        assert success is True


class TestJwtRequestLoader:
    @pytest.mark.parametrize('is_authenticated', [
        User.fake, lambda: None
    ])
    @mock.patch('keg_auth.libs.authenticators.flask_jwt_extended.verify_jwt_in_request',
                autospec=True, spec_set=True)
    @mock.patch('keg_auth.libs.authenticators.flask_jwt_extended.get_current_user',
                autospec=True, spec_set=True)
    @mock.patch('keg_auth.libs.authenticators.flask_login.login_user',
                autospec=True, spec_set=True)
    def test_user_is_authenticated(self,
                                   login_user,
                                   get_current_user,
                                   verify_jwt_in_request,
                                   is_authenticated):
        auth_user = is_authenticated()
        if not auth_user:
            verify_jwt_in_request.side_effect = flask_jwt_extended.exceptions.JWTExtendedException
        else:
            get_current_user.return_value = auth_user
        assert (auth_user is not None) == (
            auth.JwtRequestLoader.get_authenticated_user() is not None)
        if auth_user:
            login_user.assert_called_once_with(auth_user)
        else:
            assert login_user.call_count == 0

    def test_bad_token(self):
        jwt_auth = auth.JwtRequestLoader(flask.current_app)
        with mock.patch.dict(
            flask.current_app.config,
            JWT_TOKEN_LOCATION='query_string',
            JWT_QUERY_STRING_NAME='jwt',
        ):
            with flask.current_app.test_request_context('/?jwt=notgoodatall'):
                with pytest.raises(jwt.exceptions.DecodeError):
                    jwt_auth.get_authenticated_user()

    def test_missing_token(self):
        jwt_auth = auth.JwtRequestLoader(flask.current_app)
        with mock.patch.dict(
            flask.current_app.config,
            JWT_TOKEN_LOCATION='query_string',
            JWT_QUERY_STRING_NAME='jwt',
        ):
            with flask.current_app.test_request_context('/'):
                assert jwt_auth.get_authenticated_user() is None

    def test_user_not_found(self):
        user = User.fake()
        jwt_auth = auth.JwtRequestLoader(flask.current_app)
        token = jwt_auth.create_access_token(user)
        User.delete_cascaded()
        with mock.patch.dict(
            flask.current_app.config,
            JWT_TOKEN_LOCATION='query_string',
            JWT_QUERY_STRING_NAME='jwt',
        ):
            with flask.current_app.test_request_context('/?jwt={}'.format(token)):
                assert jwt_auth.get_authenticated_user() is None

    def test_user_not_active(self):
        user = User.fake(is_enabled=False)
        jwt_auth = auth.JwtRequestLoader(flask.current_app)
        token = jwt_auth.create_access_token(user)
        with mock.patch.dict(
            flask.current_app.config,
            JWT_TOKEN_LOCATION='query_string',
            JWT_QUERY_STRING_NAME='jwt',
        ):
            with flask.current_app.test_request_context('/?jwt={}'.format(token)):
                assert jwt_auth.get_authenticated_user() is None

    def test_user_verified(self):
        user = User.fake()
        jwt_auth = auth.JwtRequestLoader(flask.current_app)
        token = jwt_auth.create_access_token(user)
        with mock.patch.dict(
            flask.current_app.config,
            JWT_TOKEN_LOCATION='query_string',
            JWT_QUERY_STRING_NAME='jwt',
        ):
            with flask.current_app.test_request_context('/?jwt={}'.format(token)):
                assert jwt_auth.get_authenticated_user() is user

    def test_create_access_token(self):
        user = User.fake()
        jwt_auth = auth.JwtRequestLoader(flask.current_app)
        token = jwt_auth.create_access_token(user)
        assert flask_jwt_extended.decode_token(token)['sub'] == user.session_key


class TestPasswordPolicy:
    def setup_method(self, _):
        User.delete_cascaded()
        UserNoEmail.delete_cascaded()

    def test_check_length(self):
        user = User.fake()
        with pytest.raises(auth.PasswordPolicyError,
                           match='Password must be at least 8 characters long'):
            auth.PasswordPolicy().check_length('aBcDe1!', user)

        class LongerPolicy(auth.PasswordPolicy):
            min_length = 10

        with pytest.raises(auth.PasswordPolicyError,
                           match='Password must be at least 10 characters long'):
            LongerPolicy().check_length('aBcDeFg1!', user)

        auth.PasswordPolicy().check_length('aBcDeF1!', user)
        LongerPolicy().check_length('aBcDeFgH1!', user)

    @pytest.mark.parametrize('pw', [
        'a' * 8,
        'A' * 8,
        '1' * 8,
        'aA' * 4,
        'a1' * 4,
        '1!' * 4,
    ])
    def test_char_set_validator_failures(self, pw):
        user = User.fake()

        with pytest.raises(
            auth.PasswordPolicyError,
            match='Password must include at least 3 of lowercase letter, uppercase letter, number and/or symbol'  # noqa: E501
        ):
            auth.PasswordPolicy().check_character_set(pw, user)

    @pytest.mark.parametrize('pw', [
        'a' * 8,
        'A' * 8,
        '1' * 8,
    ])
    def test_override_min_char_types_requirement_failures(self, pw):
        user = User.fake()

        class FewerChars(auth.PasswordPolicy):
            required_min_char_types = 2

        with pytest.raises(
            auth.PasswordPolicyError,
            match='Password must include at least 2 of lowercase letter, uppercase letter, number and/or symbol'  # noqa: E501
        ):
            FewerChars().check_character_set(pw, user)

    @pytest.mark.parametrize('pw', [
        'aA' * 4,
        'A1' * 4,
        '1!' * 4,
        'aA1' * 3,
        'a1 ' * 3,
        '\t1!' * 3,
    ])
    def test_override_required_char_types_failures(self, pw):
        user = User.fake()

        class RequireWhitespace(auth.PasswordPolicy):
            required_min_char_types = 4
            required_char_types = [
                *auth.PasswordPolicy.required_char_types,
                auth.PasswordCharset('whitespace', string.whitespace)
            ]

        with pytest.raises(
            auth.PasswordPolicyError,
            match='Password must include at least 4 of lowercase letter, uppercase letter, number, symbol and/or whitespace'  # noqa: E501
        ):
            RequireWhitespace().check_character_set(pw, user)

    def test_required_char_types_one_type(self):
        user = User.fake()

        class RequireNumber(auth.PasswordPolicy):
            required_min_char_types = 1
            required_char_types = [auth.PasswordCharset('number', string.digits)]

        with pytest.raises(auth.PasswordPolicyError, match='Password must include a number'):
            RequireNumber().check_character_set('abcdefgh', user)

        RequireNumber().check_character_set('abcdefg1', user)

    @pytest.mark.parametrize('pw', [
        'aA1 ' * 3,
        'a1 !' * 3,
        '\t1!a' * 3,
    ])
    def test_override_required_char_types_success(self, pw):
        user = User.fake()

        class RequireWhitespace(auth.PasswordPolicy):
            required_min_char_types = 4
            required_char_types = [
                *auth.PasswordPolicy.required_char_types,
                auth.PasswordCharset('whitespace', string.whitespace)
            ]

        RequireWhitespace().check_character_set(pw, user)

    @pytest.mark.parametrize('pw', [
        'aaaaaaaa1!',
        'aaaaaaaaA!',
        'aaaaaaaaA1',
        'AAAAAAAA1!',
    ])
    def test_check_char_set_success(self, pw):
        user = User.fake()
        auth.PasswordPolicy().check_character_set(pw, user)

    @pytest.mark.parametrize('pw,email', [
        ('1!bob!1234', 'bob@example.com'),
        ('BoB123456!', 'bOb@example.com'),
    ])
    def test_check_does_not_contain_username_email_failures(self, pw, email):
        user = User.fake(email=email)
        with pytest.raises(auth.PasswordPolicyError, match='Password may not contain username'):
            auth.PasswordPolicy().check_does_not_contain_username(pw, user)

    @pytest.mark.parametrize('pw,username', [
        ('1!bob!1234', 'bob'),
        ('BoB123456!', 'bOb'),
    ])
    def test_check_does_not_contain_username_no_email_failures(self, pw, username):
        user = UserNoEmail.fake(username=username)
        with pytest.raises(auth.PasswordPolicyError, match='Password may not contain username'):
            auth.PasswordPolicy().check_does_not_contain_username(pw, user)

    @pytest.mark.parametrize('pw,email', [
        ('1!b0b!1234', 'bob@example.com'),
        ('B0B123456!', 'bOb@example.com'),
    ])
    def test_check_does_not_contain_username_email_success(self, pw, email):
        user = User.fake(email=email)
        auth.PasswordPolicy().check_does_not_contain_username(pw, user)

    @pytest.mark.parametrize('pw,username', [
        ('1!b0b!1234', 'bob'),
        ('B0B123456!', 'bOb'),
    ])
    def test_check_does_not_contain_username_no_email_success(self, pw, username):
        user = UserNoEmail.fake(username=username)
        auth.PasswordPolicy().check_does_not_contain_username(pw, user)



from mock import patch
import pytest

from blazeutils.strings import randchars
from keg_auth.grids import (
    ActionColumn,
    make_user_grid,
    make_group_grid,
    make_bundle_grid,
)
from webgrid import BaseGrid


class TestActionColumnLinkClasses(object):
    @pytest.mark.parametrize('grid_fun', [
        make_user_grid,
        make_group_grid,
        make_bundle_grid,
    ])
    def test_link_classes(self, grid_fun):
        class MyActionColumn(ActionColumn):
            default_view_link_class = randchars(20)
            default_edit_link_class = randchars(20)
            default_delete_link_class = randchars(20)

        class Grid(BaseGrid):
            action_column_cls = MyActionColumn

        view_class = MyActionColumn.default_view_link_class
        edit_class = MyActionColumn.default_edit_link_class
        delete_class = MyActionColumn.default_delete_link_class

        with patch.object(Grid, 'action_column_cls') as m_ac_init:
            grid_fun('edit_endpoint', None, 'delete_endpoint', None, grid_cls=Grid)
            m_ac_init.assert_called_once()

        grid = grid_fun('edit_endpoint', None, 'delete_endpoint', None, grid_cls=Grid)
        assert grid.action_column_cls == MyActionColumn

        for col in [col for col in grid.__cls_cols__ if isinstance(col, ActionColumn)]:
            assert isinstance(col, MyActionColumn)
            assert col.view_link_class_for(None) == view_class
            assert col.edit_link_class_for(None) == edit_class
            assert col.delete_link_class_for(None) == delete_class
            col.view_endpoint = 'view_endpoint'
            patch_extract_data = patch('keg_auth.grids.ActionColumn.extract_data')
            patch_grid = patch.object(col, 'grid')
            patch_url_for = patch('flask.url_for', return_value='/url')
            with patch_extract_data, patch_grid, patch_url_for:
                result = col.extract_and_format_data(None)
                for cls in [view_class, edit_class, delete_class]:
                    assert cls in result



import sys
from unittest import mock

import flask
import flask_login
import pytest

from keg_auth.libs.navigation import NavItem, NavURL

from keg_auth_ta import views

nav_menu = NavItem(
    NavItem('Home', NavURL('public.home', arg1='foo')),
    NavItem(
        'Nesting',
        NavItem('Secret1', NavURL('private.secret1')),
        NavItem('Secret1 Class', NavURL('private.secret1-class')),
    ),
    NavItem('Permissions On Stock Methods', NavURL('private.secret2')),
    NavItem('Permissions On Methods', NavURL('private.someroute')),
    NavItem('Permissions On Class And Method', NavURL('private.secret4')),
    NavItem('Permissions On NavURL',
            NavURL(
                'private.secret3', requires_permissions='permission3'
            )),
    NavItem('User Manage', NavURL('auth.user:add')),
)


class TestViewMetaInfo(object):
    @classmethod
    def setup_class(cls):
        # cause the navigation nodes to walk, assigning defining classes to the subnodes
        nav_menu.is_permitted

    def test_decorated_class_meta_user(self):
        assert views.Secret1Class.__keg_auth_requires_user__

    def test_decorated_class_meta_permissions(self):
        assert views.Secret3.__keg_auth_requires_user__
        assert views.Secret3.__keg_auth_requires_permissions__

    def test_decorated_method_meta_user(self):
        assert views.secret1.__keg_auth_requires_user__
        assert views.secret1.__keg_auth_parent_class__ is None

    def test_decorated_bound_method_meta_permissions(self):
        assert views.Secret2.get.__keg_auth_requires_user__
        assert views.Secret2.get.__keg_auth_requires_permissions__
        if sys.version_info[0] != 2:
            assert views.Secret2.get.__keg_auth_parent_class__ is views.Secret2
        else:
            assert views.Secret2.get.im_class is views.Secret2  # pragma: no cover

    def test_decorated_blueprint(self):
        assert views.protected_bp.__keg_auth_requires_user__
        assert views.protected_bp.__keg_auth_requires_permissions__


class TestNavItem(object):
    """ Test node permission logic

        Tests of node permissions are user-oriented, so we have to run these in a request context
    """

    def setup_method(self):
        self.Permission = flask.current_app.auth_manager.entity_registry.permission_cls
        self.Permission.delete_cascaded()

    def test_no_args(self):
        with pytest.raises(
            Exception, match='must provide a NavURL or a list of NavItems'
        ):
            NavItem()

    def test_node_invalid_endpoint(self):
        with pytest.raises(
            Exception, match='Endpoint pink_unicorns in navigation is not registered'
        ):
            NavItem('Foo', NavURL('pink_unicorns')).is_permitted

    def test_nav_group_not_assigned(self):
        node = NavItem('Foo', NavURL('public.home'))
        assert not node.nav_group

    def test_nav_group_auto_assigned(self):
        node = NavItem(
            NavItem(
                'This Beard Stays',
                NavItem('Foo2', NavURL('public.home')),
            ),
            NavItem(
                'Bar',
                NavItem('Bar2', NavURL('private.secret1')),
            )
        )
        assert node.sub_nodes[0].nav_group == 'this-beard-stays'

    def test_nav_group_manual_preserved(self):
        node = NavItem(
            NavItem(
                'Foo',
                NavItem('Foo2', NavURL('public.home')),
                nav_group='this-beard-stays'
            ),
            NavItem(
                'Bar',
                NavItem('Bar2', NavURL('private.secret1')),
            ),
        )
        assert node.sub_nodes[0].nav_group == 'this-beard-stays'

    def test_current_route_not_exists(self):
        node = NavItem('Foo', NavURL('public.home'))

        with flask.current_app.test_request_context('/some-random-route'):
            assert not node.has_current_route

    def test_current_route_not_matched(self):
        node = NavItem('Foo', NavURL('public.home'))

        with flask.current_app.test_request_context('/secret1'):
            assert not node.has_current_route

    def test_current_route_matched(self):
        node = NavItem('Foo', NavURL('public.home'))

        with flask.current_app.test_request_context('/'):
            assert node.has_current_route

    def test_current_route_matched_nested(self):
        node = NavItem(
            NavItem(
                'Foo',
                NavItem('Foo2', NavURL('public.home')),
            ),
            NavItem(
                'Bar',
                NavItem('Bar2', NavURL('private.secret1')),
            )
        )

        with flask.current_app.test_request_context('/'):
            assert node.has_current_route
            assert node.sub_nodes[0].has_current_route
            assert not node.sub_nodes[1].has_current_route

    def test_leaf_no_requirement(self):
        node = NavItem('Foo', NavURL('public.home'))

        with flask.current_app.test_request_context('/'):
            flask_login.logout_user()
            assert node.is_permitted

            user = flask.current_app.auth_manager.entity_registry.user_cls.fake()
            flask_login.login_user(user)
            node.clear_authorization(user.get_id())
            assert node.is_permitted

    def test_leaf_method_requires_user(self):
        node = NavItem('Foo', NavURL('private.secret1'))

        with flask.current_app.test_request_context('/'):
            flask_login.logout_user()
            assert not node.is_permitted

            user = flask.current_app.auth_manager.entity_registry.user_cls.fake()
            flask_login.login_user(user)
            node.clear_authorization(user.get_id())
            assert node.is_permitted

    @pytest.mark.parametrize('endpoint', ['private.secret1-class', 'private.secret1-flask-class'])
    def test_leaf_class_requires_user(self, endpoint):
        node = NavItem('Foo', NavURL(endpoint))

        with flask.current_app.test_request_context('/'):
            flask_login.logout_user()
            assert not node.is_permitted

            user = flask.current_app.auth_manager.entity_registry.user_cls.fake()
            flask_login.login_user(user)
            node.clear_authorization(user.get_id())
            assert node.is_permitted

    def test_leaf_method_requires_permissions(self):
        node = NavItem('Foo', NavURL('private.secret2'))

        with flask.current_app.test_request_context('/'):
            flask_login.logout_user()
            assert not node.is_permitted

            user = flask.current_app.auth_manager.entity_registry.user_cls.fake()
            flask_login.login_user(user)
            node.clear_authorization(user.get_id())
            assert not node.is_permitted

            perm1 = self.Permission.fake(token='permission1')
            perm2 = self.Permission.fake(token='permission2')
            user = flask.current_app.auth_manager.entity_registry.user_cls.fake(
                permissions=[perm1, perm2])
            flask_login.login_user(user)
            node.clear_authorization(user.get_id())
            assert node.is_permitted

    def test_leaf_method_requires_callable_permissions(self):
        node = NavItem('Foo', NavURL('private.secret_callable'))

        with flask.current_app.test_request_context('/'):
            flask_login.logout_user()
            assert not node.is_permitted

            user = flask.current_app.auth_manager.entity_registry.user_cls.fake()
            flask_login.login_user(user)
            node.clear_authorization(user.get_id())
            assert not node.is_permitted

            flask_login.current_user.email = 'foo@bar.baz'
            node.clear_authorization(user.get_id())
            assert node.is_permitted

    @pytest.mark.parametrize('endpoint', ['private.secret3', 'private.secret-flask'])
    def test_leaf_class_requires_permissions(self, endpoint):
        node = NavItem('Foo', NavURL(endpoint))

        with flask.current_app.test_request_context('/'):
            flask_login.logout_user()
            assert not node.is_permitted

            user = flask.current_app.auth_manager.entity_registry.user_cls.fake()
            flask_login.login_user(user)
            node.clear_authorization(user.get_id())
            assert not node.is_permitted

            perm1 = self.Permission.fake(token='permission1')
            perm2 = self.Permission.fake(token='permission2')
            user = flask.current_app.auth_manager.entity_registry.user_cls.fake(
                permissions=[perm1, perm2])
            flask_login.login_user(user)
            node.clear_authorization(user.get_id())
            assert node.is_permitted

    def test_leaf_class_requires_callable_permissions(self):
        node = NavItem('Foo', NavURL('callable_protected.callable-protected-class'))

        with flask.current_app.test_request_context('/'):
            flask_login.logout_user()
            assert not node.is_permitted

            user = flask.current_app.auth_manager.entity_registry.user_cls.fake()
            flask_login.login_user(user)
            node.clear_authorization(user.get_id())
            assert not node.is_permitted

            user.class_pass = True
            user.blueprint_pass = True
            node.clear_authorization(user.get_id())
            assert node.is_permitted

    @pytest.mark.parametrize('endpoint', ['private.secret4', 'private.secret-flask4'])
    def test_leaf_method_and_class_both_require(self, endpoint):
        node = NavItem('Foo', NavURL(endpoint))

        with flask.current_app.test_request_context('/'):
            flask_login.logout_user()
            assert not node.is_permitted

            user = flask.current_app.auth_manager.entity_registry.user_cls.fake()
            flask_login.login_user(user)
            node.clear_authorization(user.get_id())
            assert not node.is_permitted

            perm1 = self.Permission.fake(token='permission1')
            perm2 = self.Permission.fake(token='permission2')
            user = flask.current_app.auth_manager.entity_registry.user_cls.fake(
                permissions=[perm1])
            flask_login.login_user(user)
            node.clear_authorization(user.get_id())
            assert not node.is_permitted

            user = flask.current_app.auth_manager.entity_registry.user_cls.fake(
                permissions=[perm2])
            flask_login.login_user(user)
            node.clear_authorization(user.get_id())
            assert not node.is_permitted

            user = flask.current_app.auth_manager.entity_registry.user_cls.fake(
                permissions=[perm1, perm2])
            flask_login.login_user(user)
            node.clear_authorization(user.get_id())
            assert node.is_permitted

    def test_leaf_specifies_own_requirement(self):
        node = NavItem('Foo', NavURL('private.secret2', requires_permissions='permission1'))

        with flask.current_app.test_request_context('/'):
            flask_login.logout_user()
            assert not node.is_permitted

            user = flask.current_app.auth_manager.entity_registry.user_cls.fake()
            flask_login.login_user(user)
            node.clear_authorization(user.get_id())
            assert not node.is_permitted

            perm1 = self.Permission.fake(token='permission1')
            user = flask.current_app.auth_manager.entity_registry.user_cls.fake(
                permissions=[perm1])
            flask_login.login_user(user)
            node.clear_authorization(user.get_id())
            assert node.is_permitted

    def test_leaf_method_blueprint_requires_permissions(self):
        node = NavItem('Foo', NavURL('protected.protected_method'))

        with flask.current_app.test_request_context('/'):
            flask_login.logout_user()
            assert not node.is_permitted

            user = flask.current_app.auth_manager.entity_registry.user_cls.fake()
            flask_login.login_user(user)
            node.clear_authorization(user.get_id())
            assert not node.is_permitted

            perm1 = self.Permission.fake(token='permission1')
            user = flask.current_app.auth_manager.entity_registry.user_cls.fake(
                permissions=[perm1])
            flask_login.login_user(user)
            node.clear_authorization(user.get_id())
            assert node.is_permitted

    def test_leaf_class_blueprint_requires_permissions(self):
        node = NavItem('Foo', NavURL('protected.protected-class'))

        with flask.current_app.test_request_context('/'):
            flask_login.logout_user()
            assert not node.is_permitted

            user = flask.current_app.auth_manager.entity_registry.user_cls.fake()
            flask_login.login_user(user)
            node.clear_authorization(user.get_id())
            assert not node.is_permitted

            perm1 = self.Permission.fake(token='permission1')
            user = flask.current_app.auth_manager.entity_registry.user_cls.fake(
                permissions=[perm1])
            flask_login.login_user(user)
            node.clear_authorization(user.get_id())
            assert node.is_permitted

    def test_stem_requirement_from_subnode(self):
        node = NavItem('Menu', NavItem('Foo', NavURL('private.secret1-class')))

        with flask.current_app.test_request_context('/'):
            flask_login.logout_user()
            assert not node.is_permitted

            user = flask.current_app.auth_manager.entity_registry.user_cls.fake()
            flask_login.login_user(user)
            node.clear_authorization(user.get_id())
            assert node.is_permitted

    def test_stem_requirement_from_subnode_two_level(self):
        node = NavItem('Menu', NavItem('Menu2', NavItem('Foo', NavURL('private.secret1-class'))))

        with flask.current_app.test_request_context('/'):
            flask_login.logout_user()
            assert not node.is_permitted

            user = flask.current_app.auth_manager.entity_registry.user_cls.fake()
            flask_login.login_user(user)
            node.clear_authorization(user.get_id())
            assert node.is_permitted

    def test_permitted_subnodes(self):
        perm1 = self.Permission.fake(token='permission1')
        node = NavItem(
            'Menu',
            NavItem('Index', NavURL('public.home', requires_permissions='permission2')),
            NavItem(
                'Submenu',
                NavItem('Profile', NavURL('private.secret1', requires_permissions='permission1')),
                NavItem('Control Panel', NavURL('private.secret2', requires_permissions='permission2')),  # noqa
                NavItem('Accounts', NavURL('private.secret3', requires_permissions='permission1')),
            ),
            NavItem('History', NavURL('private.secret4', requires_permissions='permission2')),
        )
        with flask.current_app.test_request_context('/'):
            flask_login.logout_user()
            assert not node.permitted_sub_nodes

            user = flask.current_app.auth_manager.entity_registry.user_cls.fake(
                permissions=[perm1])
            flask_login.login_user(user)
            node.clear_authorization(user.get_id())
            assert len(node.permitted_sub_nodes) == 1
            assert node.permitted_sub_nodes[0].label == 'Submenu'

            assert len(node.permitted_sub_nodes[0].permitted_sub_nodes) == 2
            assert node.permitted_sub_nodes[0].permitted_sub_nodes[0].label == 'Profile'
            assert node.permitted_sub_nodes[0].permitted_sub_nodes[1].label == 'Accounts'

    def test_per_user_menu_items(self):
        node = NavItem('Foo', NavURL('private.secret2', requires_permissions='permission1'))

        with flask.current_app.test_request_context('/'):
            flask_login.logout_user()
            assert not node.is_permitted

            user = flask.current_app.auth_manager.entity_registry.user_cls.fake()
            flask_login.login_user(user)
            assert not node.is_permitted

            perm1 = self.Permission.fake(token='permission1')
            user = flask.current_app.auth_manager.entity_registry.user_cls.fake(
                permissions=[perm1])
            flask_login.login_user(user)
            assert node.is_permitted

    @mock.patch('keg_auth.libs.navigation.is_lazy_string', return_value=True)
    def test_lazy_string_label(self, _):
        # Pass a non-string label so is_lazy_string gets called.
        label = 12
        node = NavItem(label, NavURL('public.home'))
        assert node.label == label

    def test_logout_presence(self):
        node = NavItem('Foo', NavURL('auth.logout'))

        with flask.current_app.test_request_context('/'):
            flask_login.logout_user()
            assert not node.is_permitted

            user = flask.current_app.auth_manager.entity_registry.user_cls.fake()
            flask_login.login_user(user)
            node.clear_authorization(user.get_id())
            assert node.is_permitted

    def test_login_presence(self):
        node = NavItem('Foo', NavURL('auth.login', requires_anonymous=True))

        with flask.current_app.test_request_context('/'):
            flask_login.logout_user()
            assert node.is_permitted

            user = flask.current_app.auth_manager.entity_registry.user_cls.fake()
            flask_login.login_user(user)
            node.clear_authorization(user.get_id())
            assert not node.is_permitted



import flask
import flask_login


def get_current_user():
    """Helper to grab the authenticated user from the session.

    Trivial case is when the user is loaded in flask-login. If not, run the registered
    request loaders until we find a user.
    """
    # if flask_login has an authenticated user in session, that's who we want
    if flask_login.current_user and flask_login.current_user.is_authenticated:
        return flask_login.current_user

    # no user in session right now, so we need to run request loaders to see if any match
    user = None
    for loader in flask.current_app.auth_manager.request_loaders.values():
        user = loader.get_authenticated_user()
        if user:
            break
    if not user or not user.is_authenticated:
        return None
    return user


def get_domain_from_email(email):
    """Extract domain portion of email address."""
    parts = email.split('@')
    if len(parts) != 2:
        return None
    return parts[1]



from datetime import timedelta
from unittest import mock
from urllib.parse import quote

import arrow
import flask
import flask_webtest
import passlib
import pytest
import wrapt
from blazeutils import randchars
from blazeutils.containers import LazyDict
from keg import current_app
from keg.db import db

from keg_auth.libs.authenticators import AttemptLimitMixin
from keg_auth.model.types import AttemptType

has_attempt_skip_reason = 'no attempt model registered in entity registry'


def has_attempt_model():
    try:
        return flask.current_app.auth_manager.entity_registry.is_registered('attempt')
    except RuntimeError as exc:
        if 'application context' not in str(exc):
            raise
    return False


class AuthAttemptTests(object):
    """Tests to verify that automated attempt logging/blocking works as intended. These
    tests are included in the AuthTests class and are intended to be used in target
    applications to verify customization hasn't broken basic KegAuth functionality."""
    login_url = None
    forgot_password_url = None
    reset_password_url = None

    forgot_invalid_flashes = [('error', 'No user account matches: foo@bar.com')]
    forgot_lockout_flashes = [('error', 'Too many failed attempts.')]
    forgot_success_flashes = [
        ('success', 'Please check your email for the link to change your password.')
    ]

    login_invalid_flashes = [('error', 'Invalid password.')]
    login_lockout_flashes = [('error', 'Too many failed login attempts.')]
    login_success_flashes = [('success', 'Login successful.')]

    reset_lockout_flashes = [('error', 'Too many password reset attempts.')]
    reset_success_flashes = [
        ('success', 'Password changed.  Please use the new password to login below.')
    ]

    def setup_method(self):
        if has_attempt_model:
            self.attempt_ent.delete_cascaded()

    @classmethod
    def setup_class(cls):
        if has_attempt_model:
            cls.attempt_ent = flask.current_app.auth_manager.entity_registry.attempt_cls
        cls.client = flask_webtest.TestApp(flask.current_app)
        cls.group_ent = flask.current_app.auth_manager.entity_registry.group_cls
        cls.bundle_ent = flask.current_app.auth_manager.entity_registry.bundle_cls

    def do_login(self, client, email, password, submit_status=200):
        resp = client.get(self.login_url)
        resp.form['login_id'] = email
        resp.form['password'] = password
        return resp.form.submit(status=submit_status)

    def do_login_test(self, username, login_time, flashes, password='badpass',
                      submit_status=200, client=None):
        with mock.patch(
            'keg_auth.libs.authenticators.arrow.utcnow',
            return_value=login_time,
        ):
            resp = self.do_login(client or self.client, username, password, submit_status)
            assert resp.flashes == flashes

    @pytest.mark.parametrize('limit, timespan, lockout', [
        (3, 3600, 7200),
        (3, 7200, 300),
        (5, 300, 300),
    ])
    @pytest.mark.parametrize('create_user', [
        False,
        True,
    ])
    @pytest.mark.parametrize('view_config', [
        False,
        True,
    ])
    def test_login_attempts_blocked(self, limit, timespan, lockout, create_user, view_config):
        '''
        Test that login attempts get blocked after reaching the failed login attempt
        limit. Login attempts after the lockout period has passed (since the failed attempt
        that caused the lockout) should not be blocked.
        '''
        if not has_attempt_model():
            pytest.skip(has_attempt_skip_reason)
        config_key = lambda flag: 'LOGIN_' if flag else ''
        config_updates = lambda flag: {
            f'KEGAUTH_{config_key(flag)}ATTEMPT_LIMIT': limit,
            f'KEGAUTH_{config_key(flag)}ATTEMPT_TIMESPAN': timespan,
            f'KEGAUTH_{config_key(flag)}ATTEMPT_LOCKOUT': lockout,
        }
        with mock.patch.dict('flask.current_app.config', config_updates(view_config)):
            for key in config_updates(not view_config):
                if key in flask.current_app.config:
                    del flask.current_app.config[key]

            # We want to test blocking attempts for existing and non-existing users.
            username = 'foo@bar.com'
            invalid_flashes = [('error', 'No user account matches: foo@bar.com')]
            success_flashes = [('error', 'No user account matches: foo@bar.com')]
            if create_user:
                invalid_flashes = self.login_invalid_flashes
                success_flashes = self.login_success_flashes
                self.user_ent.fake(email=username, password='pass')

            assert self.attempt_ent.query.count() == 0

            last_attempt_time = arrow.utcnow()
            first_attempt_time = last_attempt_time + timedelta(seconds=-(timespan - 1))
            before_lockout_end = last_attempt_time + timedelta(seconds=lockout)
            after_lockout_end = last_attempt_time + timedelta(seconds=lockout + 1)

            def assert_attempt_count(attempt_count, failed_count, is_during_lockout=False):
                assert self.attempt_ent.query.filter_by(
                    user_input=username,
                    attempt_type='login',
                    is_during_lockout=is_during_lockout,
                ).count() == attempt_count
                assert self.attempt_ent.query.filter_by(
                    user_input=username,
                    attempt_type='login',
                    success=False,
                    is_during_lockout=is_during_lockout,
                ).count() == failed_count

            def do_test(login_time, flashes, password='badpass', submit_status=200):
                self.do_login_test(username, login_time, flashes, password, submit_status)

            do_test(first_attempt_time, invalid_flashes)
            assert_attempt_count(1, 1)
            assert_attempt_count(0, 0, is_during_lockout=True)
            for i in range(0, limit - 2):
                attempt_time = first_attempt_time + timedelta(seconds=i + 1)
                do_test(attempt_time, invalid_flashes)
                assert_attempt_count(i + 2, i + 2)
                assert_attempt_count(0, 0, is_during_lockout=True)

            do_test(last_attempt_time, invalid_flashes)
            assert_attempt_count(limit, limit)
            assert_attempt_count(0, 0, is_during_lockout=True)

            # Test attempts blocked at start of lockout.
            do_test(last_attempt_time + timedelta(seconds=1), self.login_lockout_flashes, 'pass')
            assert_attempt_count(limit, limit)
            assert_attempt_count(1, 1, is_during_lockout=True)

            # Test attempts blocked just before end of lockout.
            for i in range(0, limit):
                attempt_time = before_lockout_end - timedelta(seconds=i + 1)
                do_test(attempt_time, self.login_lockout_flashes)
                assert_attempt_count(limit, limit)
                assert_attempt_count(2 + i, i + 2, is_during_lockout=True)

            # Test attempts not blocked after lockout. Note that even though in the
            # previous loop we attempted (limit) times unsuccessfully, those attempts
            # do not count against the limit counter because they were done during
            # lockout.
            status = 302 if create_user else 200
            fail_count = limit + (0 if create_user else 1)
            do_test(after_lockout_end, success_flashes, 'pass', status)
            assert_attempt_count(limit + 1, fail_count)
            assert_attempt_count(limit + 1, limit + 1, is_during_lockout=True)

    @mock.patch.dict('flask.current_app.config', {
        'KEGAUTH_ATTEMPT_LIMIT_ENABLED': False,
        'KEGAUTH_LOGIN_ATTEMPT_LIMIT': 3,
        'KEGAUTH_LOGIN_ATTEMPT_TIMESPAN': 3600,
        'KEGAUTH_LOGIN_ATTEMPT_LOCKOUT': 7200,
    })
    def test_login_attempts_not_blocked(self):
        '''
        Test that we do not block any attempts with missing attempt entity.
        '''
        if not has_attempt_model():
            pytest.skip(has_attempt_skip_reason)
        user = self.user_ent.fake(email='foo@bar.com', password='pass')
        assert self.attempt_ent.query.count() == 0

        def do_test(attempt_count, flashes, password='badpass', submit_status=200):
            resp = self.do_login(self.client, user.email, password, submit_status)
            assert self.attempt_ent.query.filter_by(
                user_input=user.email, attempt_type='login').count() == attempt_count
            assert resp.flashes == flashes

        do_test(0, self.login_invalid_flashes)
        do_test(0, self.login_invalid_flashes)
        do_test(0, self.login_invalid_flashes)
        do_test(0, self.login_invalid_flashes)
        do_test(0, self.login_success_flashes, 'pass', 302)

    @mock.patch('flask.current_app.auth_manager.entity_registry._attempt_cls',
                new_callable=mock.PropertyMock(return_value=None))
    def test_login_attempts_blocked_but_not_configured(self, _):
        '''
        Test that we check for the registered attempt entity when limiting is enabled.
        '''
        user = self.user_ent.fake(email='foo@bar.com', password='pass')
        assert self.attempt_ent.query.count() == 0

        with pytest.raises(Exception, match=r'.*attempt entity is not registered.*'):
            self.do_login(self.client, user.email, 'badpass', 200)

    @pytest.mark.parametrize('limit, timespan, lockout', [
        (3, 3600, 7200),
        (3, 7200, 300),
        (5, 300, 300),
    ])
    def test_successful_login_resets_attempt_counter(self, limit, timespan, lockout):
        '''
        Test that several failed logins before a successful login do not count
        towards the attempt lockout counter.
        '''
        if not has_attempt_model():
            pytest.skip(has_attempt_skip_reason)
        with mock.patch.dict('flask.current_app.config', {
            'KEGAUTH_LOGIN_ATTEMPT_LIMIT': limit,
            'KEGAUTH_LOGIN_ATTEMPT_TIMESPAN': timespan,
            'KEGAUTH_LOGIN_ATTEMPT_LOCKOUT': lockout,
        }):
            user = self.user_ent.fake(email='foo@bar.com', password='pass')
            assert self.attempt_ent.query.count() == 0

            # Login and assert matching flashes and status.
            def do_test(login_time, flashes, password='badpass', submit_status=200):
                self.do_login_test(user.email, login_time, flashes, password, submit_status)

            login_time = arrow.utcnow()
            # Create (limit - 1) failed login attempts. The next failed login
            # would cause a lockout.
            for i in range(0, limit - 1):
                attempt_time = login_time + timedelta(seconds=-(i + 1))
                do_test(attempt_time, self.login_invalid_flashes)

            # Create a successful login to reset the attempt counter.
            do_test(login_time, self.login_success_flashes, 'pass', 302)

            self.client.get('/logout')

            # We can attempt (limit) more times after a successful login before
            # getting locked out.
            for i in range(0, limit):
                failed_login_time = login_time + timedelta(seconds=i + 1)
                do_test(failed_login_time, self.login_invalid_flashes)

            # We should be locked out now because we did (limit) failed attempts
            # after the successful attempt.
            do_test(login_time + timedelta(seconds=limit + 1), self.login_lockout_flashes)

    @pytest.mark.parametrize('limit, timespan, lockout', [
        (3, 3600, 7200),
    ])
    def test_login_attempts_blocked_by_ip(self, limit, timespan, lockout):
        '''
        Test that login attempts get blocked for an IP address
        '''
        if not has_attempt_model():
            pytest.skip(has_attempt_skip_reason)
        client = flask_webtest.TestApp(flask.current_app,
                                       extra_environ={'REMOTE_ADDR': '192.168.0.111'})

        with mock.patch.dict('flask.current_app.config', {
            'KEGAUTH_LOGIN_ATTEMPT_LIMIT': limit,
            'KEGAUTH_LOGIN_ATTEMPT_TIMESPAN': timespan,
            'KEGAUTH_LOGIN_ATTEMPT_LOCKOUT': lockout,
            'KEGAUTH_ATTEMPT_IP_LIMIT': True,
        }):
            def invalid_flashes(email):
                return [('error', f'No user account matches: {email}')]

            last_attempt_time = arrow.utcnow()
            first_attempt_time = last_attempt_time + timedelta(seconds=-(timespan - 1))

            def assert_attempt_count(attempt_count, failed_count, is_during_lockout=False):
                assert self.attempt_ent.query.filter_by(
                    attempt_type='login',
                    is_during_lockout=is_during_lockout,
                ).count() == attempt_count
                assert self.attempt_ent.query.filter_by(
                    attempt_type='login',
                    success=False,
                    is_during_lockout=is_during_lockout,
                ).count() == failed_count

            def do_test(username, login_time, flashes, submit_status=200, client=client):
                self.do_login_test(username, login_time, flashes, 'pass',
                                   submit_status, client=client)

            for i in range(0, limit):
                email = randchars() + '@foo.com'
                attempt_time = first_attempt_time + timedelta(seconds=i + 1)
                do_test(email, attempt_time, invalid_flashes(email))

            assert_attempt_count(limit, limit)
            assert_attempt_count(0, 0, is_during_lockout=True)

            # Test attempts blocked at start of lockout.
            email = randchars() + '@foo.com'
            do_test(email, last_attempt_time + timedelta(seconds=1), self.login_lockout_flashes)
            assert_attempt_count(limit, limit)
            assert_attempt_count(1, 1, is_during_lockout=True)

            # Attempt from another IP is not locked
            do_test(email, last_attempt_time + timedelta(seconds=1), invalid_flashes(email),
                    client=self.client)

    def do_forgot(self, client, email, submit_status=200):
        resp = client.get(self.forgot_password_url)
        resp.form['email'] = email
        return resp.form.submit(status=submit_status)

    def do_forgot_test(self, username, forgot_time, flashes, submit_status=200):
        with mock.patch(
            'keg_auth.libs.authenticators.arrow.utcnow',
            return_value=forgot_time,
        ):
            resp = self.do_forgot(self.client, username, submit_status)
            assert resp.flashes == flashes

    @pytest.mark.parametrize('limit, timespan, lockout', [
        (3, 3600, 7200),
        (3, 7200, 300),
        (5, 300, 300),
    ])
    def test_forgot_attempts_blocked(self, limit, timespan, lockout):
        '''
        Test that forgot attempts get blocked after reaching the failed forgot attempt
        limit. forgot attempts after the lockout period has passed (since the failed attempt
        that caused the lockout) should not be blocked.
        '''
        if not has_attempt_model():
            pytest.skip(has_attempt_skip_reason)
        with mock.patch.dict('flask.current_app.config', {
            'KEGAUTH_FORGOT_ATTEMPT_LIMIT': limit,
            'KEGAUTH_FORGOT_ATTEMPT_TIMESPAN': timespan,
            'KEGAUTH_FORGOT_ATTEMPT_LOCKOUT': lockout,
        }):
            # We want to test blocking attempts for existing and non-existing users.
            username = 'foo@bar.com'
            invalid_flashes = [('error', 'No user account matches: foo@bar.com')]
            success_flashes = [('error', 'No user account matches: foo@bar.com')]

            assert self.attempt_ent.query.count() == 0

            last_attempt_time = arrow.utcnow()
            first_attempt_time = last_attempt_time + timedelta(seconds=-(timespan - 1))
            before_lockout_end = last_attempt_time + timedelta(seconds=lockout)
            after_lockout_end = last_attempt_time + timedelta(seconds=lockout + 1)

            def assert_attempt_count(attempt_count, failed_count, is_during_lockout=False):
                assert self.attempt_ent.query.filter_by(
                    user_input=username,
                    attempt_type='forgot',
                    is_during_lockout=is_during_lockout,
                ).count() == attempt_count
                assert self.attempt_ent.query.filter_by(
                    user_input=username,
                    attempt_type='forgot',
                    success=False,
                    is_during_lockout=is_during_lockout,
                ).count() == failed_count

            def do_test(forgot_time, flashes, submit_status=200):
                self.do_forgot_test(username, forgot_time, flashes, submit_status)

            do_test(first_attempt_time, invalid_flashes)
            assert_attempt_count(1, 1)
            assert_attempt_count(0, 0, is_during_lockout=True)
            for i in range(0, limit - 2):
                attempt_time = first_attempt_time + timedelta(seconds=i + 1)
                do_test(attempt_time, invalid_flashes)
                assert_attempt_count(i + 2, i + 2)
                assert_attempt_count(0, 0, is_during_lockout=True)

            do_test(last_attempt_time, invalid_flashes)
            assert_attempt_count(limit, limit)
            assert_attempt_count(0, 0, is_during_lockout=True)

            # Test attempts blocked at start of lockout.
            do_test(last_attempt_time + timedelta(seconds=1), self.forgot_lockout_flashes)
            assert_attempt_count(limit, limit)
            assert_attempt_count(1, 1, is_during_lockout=True)

            # Test attempts blocked just before end of lockout.
            for i in range(0, limit):
                attempt_time = before_lockout_end - timedelta(seconds=i + 1)
                do_test(attempt_time, self.forgot_lockout_flashes)
                assert_attempt_count(limit, limit)
                assert_attempt_count(2 + i, i + 2, is_during_lockout=True)

            # Test attempts not blocked after lockout. Note that even though in the
            # previous loop we attempted (limit) times unsuccessfully, those attempts
            # do not count against the limit counter because they were done during
            # lockout.
            status = 200
            fail_count = limit + 1
            do_test(after_lockout_end, success_flashes, status)
            assert_attempt_count(limit + 1, fail_count)
            assert_attempt_count(limit + 1, limit + 1, is_during_lockout=True)

    @mock.patch.dict('flask.current_app.config', {
        'KEGAUTH_ATTEMPT_LIMIT_ENABLED': False,
        'KEGAUTH_FORGOT_ATTEMPT_LIMIT': 3,
        'KEGAUTH_FORGOT_ATTEMPT_TIMESPAN': 3600,
        'KEGAUTH_FORGOT_ATTEMPT_LOCKOUT': 7200,
    })
    def test_forgot_attempts_not_blocked(self):
        '''
        Test that we do not block any attempts with missing attempt entity.
        '''
        if not has_attempt_model():
            pytest.skip(has_attempt_skip_reason)
        assert self.attempt_ent.query.count() == 0

        def do_test(attempt_count, flashes, submit_status=200):
            resp = self.do_forgot(self.client, 'foo@bar.com', submit_status)
            assert self.attempt_ent.query.filter_by(
                user_input='foo@bar.com', attempt_type='forgot').count() == attempt_count
            assert resp.flashes == flashes

        do_test(0, self.forgot_invalid_flashes)
        do_test(0, self.forgot_invalid_flashes)

    @pytest.mark.parametrize('limit, timespan, lockout', [
        (3, 3600, 7200),
        (3, 7200, 300),
        (5, 300, 300),
    ])
    def test_successful_forgot_resets_attempt_counter(self, limit, timespan, lockout):
        '''
        Test that several failed forgots before a successful forgot do not count
        towards the attempt lockout counter.
        '''
        if not has_attempt_model():
            pytest.skip(has_attempt_skip_reason)
        with mock.patch.dict('flask.current_app.config', {
            'KEGAUTH_FORGOT_ATTEMPT_LIMIT': limit,
            'KEGAUTH_FORGOT_ATTEMPT_TIMESPAN': timespan,
            'KEGAUTH_FORGOT_ATTEMPT_LOCKOUT': lockout,
        }):
            assert self.attempt_ent.query.count() == 0

            # forgot and assert matching flashes and status.
            def do_test(forgot_time, flashes, submit_status=200):
                self.do_forgot_test('foo@bar.com', forgot_time, flashes, submit_status)

            forgot_time = arrow.utcnow()
            # Create (limit - 1) failed forgot attempts. The next failed forgot
            # would cause a lockout.
            for i in range(0, limit - 1):
                attempt_time = forgot_time + timedelta(seconds=-(i + 1))
                do_test(attempt_time, self.forgot_invalid_flashes)

            # Create a successful forgot to reset the attempt counter.
            user = self.user_ent.fake(email='foo@bar.com')
            self.do_forgot_test(user.email, forgot_time, self.forgot_success_flashes,
                                submit_status=302)
            self.user_ent.delete(user.id)

            # We can attempt (limit) more times after a successful forgot before
            # getting locked out.
            for i in range(0, limit):
                failed_forgot_time = forgot_time + timedelta(seconds=i + 1)
                do_test(failed_forgot_time, self.forgot_invalid_flashes)

            # We should be locked out now because we did (limit) failed attempts
            # after the successful attempt.
            do_test(forgot_time + timedelta(seconds=limit + 1), self.forgot_lockout_flashes)

    def do_reset_test(self, user, reset_time, flashes, submit_status=200):
        with mock.patch(
            'keg_auth.libs.authenticators.arrow.utcnow',
            return_value=reset_time,
        ):
            db.session.expire(user)
            token = user.token_generate()
            url = self.reset_password_url.format(user_id=user.id, token=token)
            resp = self.client.get(url, status=200)
            new_pass = randchars(8)
            resp.form['password'] = new_pass
            resp.form['confirm'] = new_pass
            resp = resp.form.submit(status=submit_status)
            assert resp.flashes == flashes

    @pytest.mark.parametrize('limit, timespan, lockout', [
        (3, 3600, 7200),
        (3, 7200, 300),
        (5, 300, 300),
    ])
    def test_reset_pw_attempts_blocked(self, limit, timespan, lockout):
        '''
        Test that login attempts get blocked after reaching the failed login attempt
        limit. Login attempts after the lockout period has passed (since the failed attempt
        that caused the lockout) should not be blocked.
        '''
        if not has_attempt_model():
            pytest.skip(has_attempt_skip_reason)
        with mock.patch.dict('flask.current_app.config', {
            'KEGAUTH_RESET_ATTEMPT_LIMIT': limit,
            'KEGAUTH_RESET_ATTEMPT_TIMESPAN': timespan,
            'KEGAUTH_RESET_ATTEMPT_LOCKOUT': lockout,
        }):
            user = self.user_ent.fake(email='foo@bar.com', password='pass')
            assert self.attempt_ent.query.count() == 0

            last_attempt_time = arrow.utcnow()
            first_attempt_time = last_attempt_time + timedelta(seconds=-(timespan - 1))
            before_lockout_end = last_attempt_time + timedelta(seconds=lockout)
            after_lockout_end = last_attempt_time + timedelta(seconds=lockout + 1)

            def assert_attempt_count(attempt_count, failed_count, is_during_lockout=False):
                assert self.attempt_ent.query.filter_by(
                    user_input=user.email,
                    attempt_type='reset',
                    is_during_lockout=is_during_lockout,
                ).count() == attempt_count
                assert self.attempt_ent.query.filter_by(
                    user_input=user.email,
                    attempt_type='reset',
                    success=False,
                    is_during_lockout=is_during_lockout,
                ).count() == failed_count

            def do_test(reset_time, flashes, submit_status=200):
                self.do_reset_test(user, reset_time, flashes, submit_status)

            do_test(first_attempt_time, self.reset_success_flashes, 302)
            assert_attempt_count(1, 0)
            assert_attempt_count(0, 0, is_during_lockout=True)
            for i in range(0, limit - 2):
                attempt_time = first_attempt_time + timedelta(seconds=i + 1)
                do_test(attempt_time, self.reset_success_flashes, 302)
                assert_attempt_count(i + 2, 0)
                assert_attempt_count(0, 0, is_during_lockout=True)

            do_test(last_attempt_time, self.reset_success_flashes, 302)
            assert_attempt_count(limit, 0)
            assert_attempt_count(0, 0, is_during_lockout=True)

            # Test attempts blocked at start of lockout.
            do_test(last_attempt_time + timedelta(seconds=1), self.reset_lockout_flashes)
            assert_attempt_count(limit, 0)
            assert_attempt_count(1, 1, is_during_lockout=True)

            # Test attempts blocked just before end of lockout.
            for i in range(0, limit):
                attempt_time = before_lockout_end - timedelta(seconds=i + 1)
                do_test(attempt_time, self.reset_lockout_flashes)
                assert_attempt_count(limit, 0)
                assert_attempt_count(2 + i, 2 + i, is_during_lockout=True)

            # Test attempts not blocked after lockout. Note that even though in the
            # previous loop we attempted (limit) times unsuccessfully, those attempts
            # do not count against the limit counter because they were done during
            # lockout.
            do_test(after_lockout_end, self.reset_success_flashes, 302)
            assert_attempt_count(limit + 1, 0)
            assert_attempt_count(limit + 1, limit + 1, is_during_lockout=True)

    @mock.patch.dict('flask.current_app.config', {
        'KEGAUTH_ATTEMPT_LIMIT_ENABLED': False,
        'KEGAUTH_RESET_ATTEMPT_LIMIT': 2,
        'KEGAUTH_RESET_ATTEMPT_TIMESPAN': 3600,
        'KEGAUTH_RESET_ATTEMPT_LOCKOUT': 7200,
    })
    def test_reset_pw_attempts_not_blocked(self):
        if not has_attempt_model():
            pytest.skip(has_attempt_skip_reason)
        user = self.user_ent.fake()
        assert self.attempt_ent.query.count() == 0

        def do_test(reset_time, flashes, submit_status=200):
            self.do_reset_test(user, reset_time, flashes, submit_status)

        do_test(arrow.utcnow(), self.reset_success_flashes, 302)
        do_test(arrow.utcnow(), self.reset_success_flashes, 302)
        do_test(arrow.utcnow(), self.reset_success_flashes, 302)
        do_test(arrow.utcnow(), self.reset_success_flashes, 302)

    @mock.patch('keg_auth.libs.authenticators.AttemptLimitMixin.get_request_remote_addr',
                return_value='12.12.12.12')
    def test_logs_attempt_source_ip(self, m_get_remote_addr):
        if not has_attempt_model():
            pytest.skip(has_attempt_skip_reason)
        user = self.user_ent.fake(email='foo@bar.com', password='pass')
        self.do_login(self.client, user.email, 'pass', 302)

        assert self.attempt_ent.query.one().source_ip == m_get_remote_addr.return_value

    def test_csrf_failure_logged(self):
        if not has_attempt_model():
            pytest.skip(has_attempt_skip_reason)
        resp = self.client.get(self.login_url)
        resp.form['login_id'] = 'foo@bar.baz'
        resp.form['password'] = 'notmypassword'
        resp.form['csrf_token'] = 'foo'
        resp.form.submit(status=400)

        result = self.attempt_ent.query.one()
        assert result.attempt_type == AttemptType.login
        assert not result.success
        assert result.user_input == 'foo@bar.baz'

    def test_password_form_validation_failure_logged(self):
        if not has_attempt_model():
            pytest.skip(has_attempt_skip_reason)
        resp = self.client.get(self.login_url)
        resp.form['login_id'] = 'myinputisnotanemail'
        resp.form['password'] = 'notmypassword'
        resp.form.submit(status=200)

        result = self.attempt_ent.query.one()
        assert result.attempt_type == AttemptType.login
        assert not result.success
        assert result.user_input == 'myinputisnotanemail'

    def test_forgot_form_validation_failure_logged(self):
        if not has_attempt_model():
            pytest.skip(has_attempt_skip_reason)
        resp = self.client.get(self.forgot_password_url)
        resp.form['email'] = 'myinputisnotanemail'
        resp.form.submit(status=200)

        result = self.attempt_ent.query.one()
        assert result.attempt_type == AttemptType.forgot
        assert not result.success
        assert result.user_input == 'myinputisnotanemail'

    def test_get_request_remote_addr(self):
        if not has_attempt_model():
            pytest.skip(has_attempt_skip_reason)
        with current_app.test_request_context(environ_base={'REMOTE_ADDR': '12.12.12.12'}):
            assert AttemptLimitMixin.get_request_remote_addr() == '12.12.12.12'


class AuthTests(AuthAttemptTests):
    """
        These tests are designed so they can can be imported into an application's tests
        and ran to ensure customization of KegAuth hasn't broken basic functionality.
    """
    login_url = '/login'
    forgot_password_url = '/forgot-password'
    reset_password_url = '/reset-password/{user_id}/{token}'
    after_login_url = '/'
    logout_url = '/logout'
    after_logout_url = login_url
    protected_url = '/secret1'
    protected_url_permissions = None

    def setup_method(self):
        super().setup_method()
        self.user_ent.delete_cascaded()

    def test_login_get(self):
        app = flask.current_app
        client = flask_webtest.TestApp(app)
        resp = client.get(self.login_url)
        assert resp.status_code == 200

    def test_login_head(self):
        client = flask_webtest.TestApp(flask.current_app)
        client.head(self.login_url, status=405)

    def test_login_form_error(self):
        client = flask_webtest.TestApp(flask.current_app)
        resp = client.get(self.login_url)

        resp.form['login_id'] = 'foo'
        resp = resp.form.submit(status=200)

        flash_form_error = flask.current_app.auth_manager.login_authenticator_cls.\
            responder_cls['login'].flash_form_error
        category = flash_form_error[1]
        message = flash_form_error[0]
        assert resp.flashes == [(category, message)]

    def test_login_field_success(self):
        user = self.user_ent.fake(email='foo@bar.com', password='pass')

        client = flask_webtest.TestApp(flask.current_app)
        resp = client.get(self.login_url)

        resp.form['login_id'] = 'foo@bar.com'
        resp.form['password'] = 'pass'
        resp = resp.form.submit()

        assert resp.status_code == 302, resp.html
        assert resp.headers['Location'] == self.after_login_url
        flash_success = flask.current_app.auth_manager.login_authenticator_cls.responder_cls['login'].flash_success  # noqa
        category = flash_success[1]
        message = flash_success[0]
        assert resp.flashes == [(category, message)]
        if has_attempt_model:
            assert self.attempt_ent.query.count() == 1
            assert self.attempt_ent.query.filter_by(
                attempt_type='login',
                success=True,
                user_input=user.email,
                is_during_lockout=False,
            )

    def test_login_field_success_next_parameter(self):
        self.user_ent.fake(email='foo@bar.com', password='pass')

        next = '/foo'
        client = flask_webtest.TestApp(flask.current_app)
        resp = client.get('{}?next={}'.format(self.login_url, next))

        resp.form['login_id'] = 'foo@bar.com'
        resp.form['password'] = 'pass'
        resp = resp.form.submit()

        assert resp.status_code == 302, resp.html
        assert resp.headers['Location'] == next
        flash_success = flask.current_app.auth_manager.login_authenticator_cls.responder_cls['login'].flash_success  # noqa
        category = flash_success[1]
        message = flash_success[0]
        assert resp.flashes == [(category, message)]

    def test_login_field_success_next_session(self):
        self.user_ent.fake(email='foo@bar.com', password='pass')

        next = '/foo'
        with mock.patch.dict(flask.current_app.config, {'USE_SESSION_FOR_NEXT': True}):
            client = flask_webtest.TestApp(flask.current_app)
            with client.session_transaction() as sess:
                sess['next'] = next
            resp = client.get(self.login_url)

            resp.form['login_id'] = 'foo@bar.com'
            resp.form['password'] = 'pass'
            resp = resp.form.submit()

        assert resp.status_code == 302, resp.html
        assert resp.headers['Location'] == next
        flash_success = flask.current_app.auth_manager.login_authenticator_cls.responder_cls['login'].flash_success  # noqa
        category = flash_success[1]
        message = flash_success[0]
        assert resp.flashes == [(category, message)]

    def test_next_parameter_not_open_redirect(self):
        """ensure following the "next" parameter doesn't allow for an open redirect"""
        self.user_ent.fake(email='foo@bar.com', password='pass')

        # unquoted next parameter
        next = 'http://www.example.com'
        client = flask_webtest.TestApp(flask.current_app)
        resp = client.get('{}?next={}'.format(self.login_url, next))

        resp.form['login_id'] = 'foo@bar.com'
        resp.form['password'] = 'pass'
        resp = resp.form.submit()

        assert resp.status_code == 302, resp.html
        # verify the 'next' parameter was ignored
        assert resp.headers['Location'] == self.after_login_url
        flash_success = flask.current_app.auth_manager.login_authenticator_cls.responder_cls['login'].flash_success  # noqa
        category = flash_success[1]
        message = flash_success[0]
        assert resp.flashes == [(category, message)]

        # quoted next parameter
        client = flask_webtest.TestApp(flask.current_app)
        resp = client.get('{}?next={}'.format(self.login_url, quote(next)))

        resp.form['login_id'] = 'foo@bar.com'
        resp.form['password'] = 'pass'
        resp = resp.form.submit()

        assert resp.status_code == 302, resp.html
        # verify the 'next' parameter was ignored
        assert resp.headers['Location'] == self.after_login_url
        flash_success = flask.current_app.auth_manager.login_authenticator_cls.responder_cls['login'].flash_success  # noqa
        category = flash_success[1]
        message = flash_success[0]
        assert resp.flashes == [(category, message)]

    def test_login_invalid_password(self):
        user = self.user_ent.fake(email='foo@bar.com', password='pass')
        if has_attempt_model:
            assert self.attempt_ent.query.count() == 0

        client = flask_webtest.TestApp(flask.current_app)
        resp = client.get(self.login_url)

        resp.form['login_id'] = 'foo@bar.com'
        resp.form['password'] = 'badpass'
        resp = resp.form.submit(status=200)

        flash_invalid_password = flask.current_app.auth_manager.login_authenticator_cls.\
            responder_cls['login'].flash_invalid_password
        category = flash_invalid_password[1]
        message = flash_invalid_password[0]
        assert resp.flashes == [(category, message)]
        if has_attempt_model:
            assert self.attempt_ent.query.count() == 1
            assert self.attempt_ent.get_by(
                attempt_type='login',
                success=False,
                user_input=user.email,
                is_during_lockout=False,
            )

    def test_login_user_missing(self):
        client = flask_webtest.TestApp(flask.current_app)
        resp = client.get(self.login_url)

        resp.form['login_id'] = 'foo@bar.com'
        resp.form['password'] = 'badpass'
        resp = resp.form.submit(status=200)

        flash_invalid_user = flask.current_app.auth_manager.login_authenticator_cls.\
            responder_cls['login'].flash_invalid_user
        category = flash_invalid_user[1]
        message = flash_invalid_user[0]
        assert resp.flashes == [(category, message.format('foo@bar.com'))]

    def test_login_user_unverified(self):
        self.user_ent.fake(email='foo@bar.com', password='pass', is_verified=False)

        client = flask_webtest.TestApp(flask.current_app)
        resp = client.get(self.login_url)

        resp.form['login_id'] = 'foo@bar.com'
        resp.form['password'] = 'badpass'
        resp = resp.form.submit(status=200)

        flash_unverified_user = flask.current_app.auth_manager.login_authenticator_cls.\
            responder_cls['login'].flash_unverified_user
        category = flash_unverified_user[1]
        message = flash_unverified_user[0]
        assert resp.flashes == [(category, message.format('foo@bar.com'))]

    def test_login_user_disabled(self):
        self.user_ent.fake(email='foo@bar.com', password='pass', is_enabled=False)

        client = flask_webtest.TestApp(flask.current_app)
        resp = client.get(self.login_url)

        resp.form['login_id'] = 'foo@bar.com'
        resp.form['password'] = 'badpass'
        resp = resp.form.submit(status=200)

        flash_disabled_user = flask.current_app.auth_manager.login_authenticator_cls.\
            responder_cls['login'].flash_disabled_user
        category = flash_disabled_user[1]
        message = flash_disabled_user[0]
        assert resp.flashes == [(category, message.format('foo@bar.com'))]

    def test_login_user_disabled_timestamp(self):
        self.user_ent.fake(
            email='foo@bar.com',
            password='pass',
            disabled_utc=arrow.utcnow().shift(days=-10)
        )

        client = flask_webtest.TestApp(flask.current_app)
        resp = client.get(self.login_url)

        resp.form['login_id'] = 'foo@bar.com'
        resp.form['password'] = 'badpass'
        resp = resp.form.submit(status=200)

        flash_disabled_user = flask.current_app.auth_manager.login_authenticator_cls. \
            responder_cls['login'].flash_disabled_user
        category = flash_disabled_user[1]
        message = flash_disabled_user[0]
        assert resp.flashes == [(category, message.format('foo@bar.com'))]

    def test_login_protection(self):
        self.user_ent.fake(
            email='foo@bar.com', password='pass', permissions=self.protected_url_permissions
        )

        client = flask_webtest.TestApp(flask.current_app)
        resp = client.get(self.protected_url, status=302)
        assert resp.headers['Location'].startswith(self.login_url)

        resp = resp.follow()
        resp.form['login_id'] = 'foo@bar.com'
        resp.form['password'] = 'pass'
        resp = resp.form.submit(status=302)
        flash_success = flask.current_app.auth_manager.login_authenticator_cls.responder_cls['login'].flash_success  # noqa
        category = flash_success[1]
        message = flash_success[0]
        assert resp.flashes == [(category, message)]

        # Now that we have logged in, we should be able to get to the page.
        client.get(self.protected_url, status=200)

    def test_forgot_pw_form_error(self):
        client = flask_webtest.TestApp(flask.current_app)
        resp = client.get(self.forgot_password_url)
        resp = resp.form.submit(status=200)

        flash_form_error = flask.current_app.auth_manager.login_authenticator_cls.\
            responder_cls['forgot-password'].flash_form_error
        category = flash_form_error[1]
        message = flash_form_error[0]
        assert resp.flashes == [(category, message)]

    def test_forgot_pw_invalid_user(self):
        client = flask_webtest.TestApp(flask.current_app)
        resp = client.get(self.forgot_password_url)

        resp.form['email'] = 'foo@bar.com'
        resp = resp.form.submit(status=200)

        flash_invalid_user = flask.current_app.auth_manager.login_authenticator_cls.\
            responder_cls['forgot-password'].flash_invalid_user
        category = flash_invalid_user[1]
        message = flash_invalid_user[0]
        assert resp.flashes == [(category, message.format('foo@bar.com'))]

    def test_forgot_pw_user_disabled(self):
        self.user_ent.fake(email='foo@bar.com', password='pass', is_enabled=False)

        client = flask_webtest.TestApp(flask.current_app)
        resp = client.get(self.forgot_password_url)

        resp.form['email'] = 'foo@bar.com'
        resp = resp.form.submit(status=200)

        flash_disabled_user = flask.current_app.auth_manager.login_authenticator_cls.\
            responder_cls['forgot-password'].flash_disabled_user
        category = flash_disabled_user[1]
        message = flash_disabled_user[0]
        assert resp.flashes == [(category, message.format('foo@bar.com'))]

    def test_forgot_pw_success(self):
        self.user_ent.fake(email='foo@bar.com')

        client = flask_webtest.TestApp(flask.current_app)
        resp = client.get(self.forgot_password_url)

        resp.form['email'] = 'foo@bar.com'
        resp = resp.form.submit(status=302)

        flash_success = flask.current_app.auth_manager.login_authenticator_cls.\
            responder_cls['forgot-password'].flash_success
        category = flash_success[1]
        message = flash_success[0]
        assert resp.flashes == [(category, message)]

        assert resp.headers['Location'] == self.login_url

    def test_reset_pw_success(self):
        user = self.user_ent.fake()
        if has_attempt_model:
            assert self.attempt_ent.query.count() == 0
        token = user.token_generate()
        url = self.reset_password_url.format(user_id=user.id, token=token)

        client = flask_webtest.TestApp(flask.current_app)
        resp = client.get(url, status=200)

        resp.form['password'] = 'fooBar123'
        resp.form['confirm'] = 'fooBar123'
        resp = resp.form.submit(status=302)

        flash_success = flask.current_app.auth_manager.login_authenticator_cls.\
            responder_cls['reset-password'].flash_success
        category = flash_success[1]
        message = flash_success[0]
        assert resp.flashes == [(category, message)]

        assert resp.headers['Location'] == self.login_url
        if has_attempt_model:
            assert self.attempt_ent.query.count() == 1
            assert self.attempt_ent.get_by(attempt_type='reset', user_input=user.email)

    def test_reset_pw_form_error(self):
        user = self.user_ent.fake()
        token = user.token_generate()
        url = self.reset_password_url.format(user_id=user.id, token=token)

        client = flask_webtest.TestApp(flask.current_app)
        resp = client.get(url, status=200)
        resp = resp.form.submit(status=200)

        flash_form_error = flask.current_app.auth_manager.login_authenticator_cls.\
            responder_cls['reset-password'].flash_form_error
        category = flash_form_error[1]
        message = flash_form_error[0]
        assert resp.flashes == [(category, message)]

    def test_reset_pw_missing_user(self):
        url = self.reset_password_url.format(user_id='99999999', token='123')

        client = flask_webtest.TestApp(flask.current_app)
        client.get(url, status=404)

    def test_reset_pw_bad_token(self):
        user = self.user_ent.fake()
        url = url = self.reset_password_url.format(user_id=user.id, token='abc')

        client = flask_webtest.TestApp(flask.current_app)
        resp = client.get(url, status=302)

        flash_invalid_token = flask.current_app.auth_manager.login_authenticator_cls.\
            responder_cls['reset-password'].flash_invalid_token
        category = flash_invalid_token[1]
        message = flash_invalid_token[0]
        assert resp.flashes == [(category, message)]

        assert resp.headers['Location'] == self.forgot_password_url

    def test_verify_account_success(self):
        user = self.user_ent.fake(is_verified=False)
        assert not user.is_verified

        user.token_generate()
        url = flask.current_app.auth_manager.mail_manager.verify_account_url(user)

        client = flask_webtest.TestApp(flask.current_app)
        resp = client.get(url, status=200)

        resp.form['password'] = 'fooBar123'
        resp.form['confirm'] = 'fooBar123'
        resp = resp.form.submit(status=302)

        flash_success = flask.current_app.auth_manager.login_authenticator_cls.\
            responder_cls['verify-account'].flash_success
        category = flash_success[1]
        message = flash_success[0]
        assert resp.flashes == [(category, message)]

        assert resp.headers['Location'] == self.login_url

        assert self.user_ent.query.filter_by(id=user.id, is_verified=True).count() == 1

    def test_verify_account_form_error(self):
        user = self.user_ent.fake()
        user.token_generate()
        url = flask.current_app.auth_manager.mail_manager.verify_account_url(user)

        client = flask_webtest.TestApp(flask.current_app)
        resp = client.get(url, status=200)
        resp = resp.form.submit(status=200)

        flash_form_error = flask.current_app.auth_manager.login_authenticator_cls.\
            responder_cls['verify-account'].flash_form_error
        category = flash_form_error[1]
        message = flash_form_error[0]
        assert resp.flashes == [(category, message)]

    def test_verify_account_missing_user(self):
        user = LazyDict(id=9999999, _token_plain='123')
        url = flask.current_app.auth_manager.mail_manager.verify_account_url(user)

        client = flask_webtest.TestApp(flask.current_app)
        client.get(url, status=404)

    def test_verify_account_bad_token(self):
        user = self.user_ent.fake()
        user._token_plain = 'abc'
        url = flask.current_app.auth_manager.mail_manager.verify_account_url(user)

        client = flask_webtest.TestApp(flask.current_app)
        resp = client.get(url, status=302)

        flash_invalid_token = flask.current_app.auth_manager.login_authenticator_cls.\
            responder_cls['verify-account'].flash_invalid_token
        category = flash_invalid_token[1]
        message = flash_invalid_token[0]
        assert resp.flashes == [(category, message)]

        assert resp.headers['Location'] == self.forgot_password_url

    def test_logout(self):
        user = self.user_ent.fake(permissions=self.protected_url_permissions)
        client = flask_webtest.TestApp(flask.current_app)
        with client.session_transaction() as sess:
            sess['_user_id'] = user.session_key

        # Make sure our client is actually logged in
        client.get(self.protected_url, status=200)

        # logout
        resp = client.get(self.logout_url, status=302)
        assert resp.flashes == [('success', 'You have been logged out.')]

        # Check redirect location
        assert resp.headers['Location'] == self.after_logout_url

        # Confirm logout occured
        client.get(self.protected_url, status=302)

    def test_list_user(self):
        user = self.user_ent.fake(permissions=['auth-manage'])
        client = AuthTestApp(flask.current_app, user=user)
        resp = client.get('/users?op(username)=eq&v1(username)=' + user.email)
        assert resp.pyquery('.grid-header-add-link a').attr('href').startswith('/users/add')
        assert resp.pyquery('h1').eq(0).text() == 'Users'
        assert resp.pyquery('.grid-header-add-link').eq(0).text() == 'Create User'

    def test_edit_user(self):
        user_edit = self.user_ent.fake()
        user = self.user_ent.fake(permissions=['auth-manage'])
        client = AuthTestApp(flask.current_app, user=user)
        resp = client.get('/users/{}/edit'.format(user_edit.id))
        assert resp.pyquery('h1').eq(0).text() == 'Edit User'
        resp = resp.form.submit()

        assert resp.status_code == 302
        assert resp.location.endswith('/users')
        assert resp.flashes == [('success', 'Successfully modified User')]

    def test_delete_user(self):
        user_delete_id = self.user_ent.fake().id
        user = self.user_ent.fake(permissions=['auth-manage'])
        client = AuthTestApp(flask.current_app, user=user)
        resp = client.get('/users/{}/delete'.format(user_delete_id))

        assert resp.status_code == 302
        assert resp.location.endswith('/users')
        assert resp.flashes == [('success', 'Successfully removed User')]

        assert not self.user_ent.query.get(user_delete_id)

    def test_list_group(self):
        user = self.user_ent.fake(permissions=['auth-manage'])
        client = AuthTestApp(flask.current_app, user=user)
        group = self.group_ent.fake()
        resp = client.get('/groups?op(name)=eq&v1(name)=' + group.name)
        assert resp.pyquery('.grid-header-add-link a').attr('href').startswith('/groups/add')
        assert resp.pyquery('h1').eq(0).text() == 'Groups'
        assert resp.pyquery('.grid-header-add-link').eq(0).text() == 'Create Group'

    def test_add_group(self):
        user = self.user_ent.fake(permissions=['auth-manage'])
        client = AuthTestApp(flask.current_app, user=user)
        resp = client.get('/groups/add')
        group_name = randchars()
        resp.form['name'] = group_name
        resp = resp.form.submit()
        assert resp.status_code == 302
        assert resp.location.endswith('/groups')
        assert resp.flashes == [('success', 'Successfully created Group')]

        assert self.group_ent.get_by(name=group_name)

    def test_edit_group(self):
        user = self.user_ent.fake(permissions=['auth-manage'])
        client = AuthTestApp(flask.current_app, user=user)
        group = self.group_ent.fake()
        resp = client.get(f'/groups/{group.id}/edit')
        resp = resp.form.submit()
        assert resp.status_code == 302
        assert resp.location.endswith('/groups')
        assert resp.flashes == [('success', 'Successfully modified Group')]

    def test_delete_group(self):
        group_delete_id = self.group_ent.fake().id
        user = self.user_ent.fake(permissions=['auth-manage'])
        client = AuthTestApp(flask.current_app, user=user)
        resp = client.get(f'/groups/{group_delete_id}/delete')

        assert resp.status_code == 302
        assert resp.location.endswith('/groups')
        assert resp.flashes == [('success', 'Successfully removed Group')]

        assert not self.group_ent.query.get(group_delete_id)

    def test_list_bundle(self):
        user = self.user_ent.fake(permissions=['auth-manage'])
        client = AuthTestApp(flask.current_app, user=user)
        bundle = self.bundle_ent.fake()
        resp = client.get('/bundles?op(name)=eq&v1(name)=' + bundle.name)
        assert resp.pyquery('.grid-header-add-link a').attr('href').startswith('/bundles/add')
        assert resp.pyquery('h1').eq(0).text() == 'Bundles'
        assert resp.pyquery('.grid-header-add-link').eq(0).text() == 'Create Bundle'

    def test_add_bundle(self):
        user = self.user_ent.fake(permissions=['auth-manage'])
        client = AuthTestApp(flask.current_app, user=user)
        resp = client.get('/bundles/add')
        bundle_name = randchars()
        resp.form['name'] = bundle_name
        resp = resp.form.submit()
        assert resp.status_code == 302
        assert resp.location.endswith('/bundles')
        assert resp.flashes == [('success', 'Successfully created Bundle')]

        assert self.bundle_ent.get_by(name=bundle_name)

    def test_edit_bundle(self):
        user = self.user_ent.fake(permissions=['auth-manage'])
        client = AuthTestApp(flask.current_app, user=user)
        bundle = self.bundle_ent.fake()
        resp = client.get(f'/bundles/{bundle.id}/edit')
        resp = resp.form.submit()
        assert resp.status_code == 302
        assert resp.location.endswith('/bundles')
        assert resp.flashes == [('success', 'Successfully modified Bundle')]

    def test_delete_bundle(self):
        bundle_delete_id = self.bundle_ent.fake().id
        user = self.user_ent.fake(permissions=['auth-manage'])
        client = AuthTestApp(flask.current_app, user=user)
        resp = client.get(f'/bundles/{bundle_delete_id}/delete')

        assert resp.status_code == 302
        assert resp.location.endswith('/bundles')
        assert resp.flashes == [('success', 'Successfully removed Bundle')]

        assert not self.bundle_ent.query.get(bundle_delete_id)


@wrapt.decorator
def user_request(wrapped, instance, args, kwargs):
    new_kwargs = kwargs.copy()
    user = new_kwargs.pop('user', None)
    extra_environ = new_kwargs.setdefault('extra_environ', {})
    if user is not None:
        extra_environ['TEST_USER_ID'] = str(user.session_key)
    return wrapped(*args, **new_kwargs)


def with_crypto_context(field, context=None):
    """Wrap a test to use a real cryptographic context for a :class:`KAPasswordType`

    Temporarily assign a :class:`passlib.context.CryptoContext` to a particular entity column.

    :param context (optional): :class:`passlib.context.CryptoContext` to use for this test. The
        default value is `keg_auth.core.DEFAULT_CRYPTO_SCHEMES`.

    .. NOTE:

    In most situations we don't want a real crypto scheme to run in the tests, it is
    slow on entities like Users which have a password. ``User.fake`` will generate a value
    for that instance and then hash which takes a bunch of time. However, when testing certain
    schemes, it is useful to execute the real behavior instead of the ``plaintext`` behaviour.

    ::

        import bcrypt

        bcrypt_context = passlib.context.CryptContext(scheme=['bcrypt'])

        @with_crypto_context(ents.User.password, context=bcrypt_context)
        def test_with_real_context():
            user = ents.User.fake(password='abc')
            assert bcrypt.checkpw('abc', user.password.hash)

    """
    import keg_auth

    @wrapt.decorator
    def wrapper(wrapped, instance, args, kwargs):
        prev_context = field.type.context
        field.type.context = (
            context or passlib.context.CryptContext(schemes=keg_auth.core.DEFAULT_CRYPTO_SCHEMES)
        )

        wrapped(*args, **kwargs)

        field.type.context = prev_context

    return wrapper


class AuthTestApp(flask_webtest.TestApp):
    """Wrapper of `flask_webtest.TestApp` that will inject a user into the session.

    Pass in a user instance to "log in" the session:

        user = User.fake(permissions=['auth-manage', 'do-something'])
        test_app = AuthTestApp(flask.current_app, user=user)

    When running integration tests, following the view sequence to log a user in can
    be quite time-consuming and unnecessary. Login tests can be elsewhere. Once a user
    is logged in, they are identified by their `session_key`. So, we simply inject
    that key in the environment, and then follow the request out to webtest per
    normal.
    """
    def __init__(self, app, **kwargs):
        user = kwargs.pop('user', None)
        extra_environ = kwargs.pop('extra_environ', {})
        if user is not None:
            extra_environ['TEST_USER_ID'] = str(user.session_key)
        super(AuthTestApp, self).__init__(app, extra_environ=extra_environ, **kwargs)

    @user_request
    def get(self, *args, **kwargs):
        return super(AuthTestApp, self).get(*args, **kwargs)

    @user_request
    def post(self, *args, **kwargs):
        return super(AuthTestApp, self).post(*args, **kwargs)

    @user_request
    def put(self, *args, **kwargs):
        return super(AuthTestApp, self).put(*args, **kwargs)

    @user_request
    def patch(self, *args, **kwargs):
        return super(AuthTestApp, self).patch(*args, **kwargs)

    @user_request
    def delete(self, *args, **kwargs):
        return super(AuthTestApp, self).delete(*args, **kwargs)

    @user_request
    def options(self, *args, **kwargs):
        return super(AuthTestApp, self).options(*args, **kwargs)

    @user_request
    def head(self, *args, **kwargs):
        return super(AuthTestApp, self).head(*args, **kwargs)

    @user_request
    def post_json(self, *args, **kwargs):
        return super(AuthTestApp, self).post_json(*args, **kwargs)

    @user_request
    def put_json(self, *args, **kwargs):
        return super(AuthTestApp, self).put_json(*args, **kwargs)

    @user_request
    def patch_json(self, *args, **kwargs):
        return super(AuthTestApp, self).patch_json(*args, **kwargs)

    @user_request
    def delete_json(self, *args, **kwargs):
        return super(AuthTestApp, self).delete_json(*args, **kwargs)


class ViewTestBase:
    """ Simple helper class that will set up Permission tokens as specified, log in a user, and
        provide the test app client on the class for use in tests.

        Usage: ``permissions`` class attribute can be scalar or list.

        For tests:

        - ``self.current_user``: User instance that is logged in
        - ``self.client``: AuthTestApp instance
    """
    permissions = tuple()

    @classmethod
    def create_user(cls):
        """ Creates a User record for tests. By default, simply calls ``fake`` with
            permissions."""
        return cls.user_ent.fake(permissions=cls.permissions)

    @classmethod
    def setup_user(cls):
        """Hook to do further setup on ``cls.current_user``."""
        pass

    @classmethod
    def setup_class(cls):
        cls.user_ent = flask.current_app.auth_manager.entity_registry.user_cls
        cls.permission_ent = flask.current_app.auth_manager.entity_registry.permission_cls

        # ensure all of the tokens exists
        flask.current_app.auth_manager.validate_permission_set(cls.permissions)

        cls.current_user = cls.create_user()
        cls.setup_user()
        cls.client = AuthTestApp(flask.current_app, user=cls.current_user)



import flask
import flask_login
import inflect
import keg.web
import sqlalchemy as sa
from blazeutils.strings import case_cw2dash
from keg.db import db
from flask_wtf.csrf import validate_csrf
from keg_auth import forms, grids, requires_permissions
from keg_auth.extensions import lazy_gettext as _, flash

try:
    from speaklater import is_lazy_string
except ImportError:
    is_lazy_string = lambda value: False  # noqa: E731


class CrudView(keg.web.BaseView):
    """Base CRUD view class providing add/edit/delete/list functionality.

    Basic subclass setup involves:
    - set the `grid_cls`, `form_cls`, and `orm_cls` attributes
    - set `object_name` to be the human readable label.
    - assign `object_name_plural` only if necessary
    - assign base permissions for reach of the four endpoints

    Grid is assumed to be WebGrid. Form is assumed to be WTForms. ORM is
    assumed to be SQLAlchemy. Default templates are provided with keg-auth.

    Permissions are set for each endpoint under the `permissions` dict attribute.
    Note that it is usually helpful to put a general @requires_permissions on the
    class itself, as that will aid in conditionally displaying navigation links
    based on a user's access level.
    """

    grid_cls = None
    form_cls = None
    orm_cls = None
    form_template = 'keg-auth/crud-addedit.html'
    grid_template = 'keg-auth/crud-list.html'
    object_name = None
    _inflect = inflect.engine()
    permissions = {
        'add': None,
        'edit': None,
        'delete': None,
        'list': None
    }

    @classmethod
    def map_method_route(cls, method_name, route, methods):
        method_route = keg.web.MethodRoute(method_name, route, {'methods': methods},
                                           cls.calc_url(use_blueprint=False),
                                           cls.calc_endpoint(use_blueprint=False))
        mr_options = method_route.options()
        view_func = cls.as_view(method_route.view_func_name,
                                method_route.sanitized_method_name('_'))
        cls.view_funcs[method_route.endpoint] = view_func
        mr_options['view_func'] = cls.view_funcs[method_route.endpoint]
        cls.blueprint.add_url_rule(method_route.rule(), **mr_options)

    @classmethod
    def init_routes(cls):
        """ Creates the standard set of routes from methods (add, edit, delete, list).

            To extend to further action routes:
                `cls.map_method_route(method_name, url, HTTP methods)`
                ex. `cls.map_method_route('read', '/foo', ('GET', ))`"""
        super(CrudView, cls).init_routes()

        cls.map_method_route('add', '{}/add'.format(cls.calc_url(use_blueprint=False)),
                             ('GET', 'POST'))
        cls.map_method_route(
            'edit',
            '{}/<int:objid>/edit'.format(cls.calc_url(use_blueprint=False)),
            ('GET', 'POST')
        )
        cls.map_method_route(
            'delete',
            '{}/<int:objid>/delete'.format(cls.calc_url(use_blueprint=False)),
            ('GET', )
        )
        cls.map_method_route('list', '{}'.format(cls.calc_url(use_blueprint=False)),
                             ('GET', 'POST'))

    def __init__(self, *args, **kwargs):
        super(CrudView, self).__init__(*args, **kwargs)
        self.objinst = None

    @property
    def object_name_plural(self):
        """Plural version of `object_name`. Uses the inflect library for a default value."""
        return self._inflect.plural(
            self.object_name
            if not is_lazy_string(self.object_name)
            else str(self.object_name)
        )

    def page_title(self, action):
        """Generates a heading title based on the page action.

        `action` should be a string. Values "Create" and "Edit" are handled, with a
        fall-through to return `object_name_plural` (for the list case).
        """
        if action == _('Create'):
            return _('Create {name}').format(name=self.object_name)

        if action == _('Edit'):
            return _('Edit {name}').format(name=self.object_name)

        return self.object_name_plural

    def create_form(self, obj):
        """Create an instance of `form_cls`. Must return a form if overloaded.

        `obj` is an instance of `orm_cls` (edit) or None (add).
        """
        return self.form_cls(obj=obj)

    def form_page_heading(self, action):
        """Allows customization of add/edit heading. Defaults to `page_title`."""
        return self.page_title(action)

    def form_template_args(self, arg_dict):
        """Allows customization of jinja template args for add/edit views.

        `arg_dict` contains the default arguments, including anything set with `self.assign`.

        Must return a dict of template args.
        """
        return arg_dict

    def render_form(self, obj, action, form, action_button_text=_('Save Changes')):
        """Renders the form template.

        Template arguments may be customized with the `form_template_args` method.
        """
        title_var = flask.current_app.config.get('KEGAUTH_TEMPLATE_TITLE_VAR')

        # args added with self.assign should be passed through here
        template_args = self.form_template_args(dict(self.template_args, **{
            'action': action,
            'action_button_text': action_button_text,
            'cancel_url': self.cancel_url(),
            'form': form,
            'obj_inst': obj,
            title_var: self.page_title(action),
            'page_heading': self.form_page_heading(action),
        }))
        return flask.render_template(self.form_template, **template_args)

    def add_orm_obj(self):
        """Generate a blank object instance and add it to the session."""
        o = self.orm_cls()
        db.session.add(o)
        return o

    def update_obj(self, obj, form):
        """Update an existing object instance from form data. Does not explicitly
        flush or commit."""
        obj = obj or self.add_orm_obj()
        form.populate_obj(obj)
        return obj

    def add_edit(self, meth, obj=None):
        """Handle form-related requests for add/edit.

        Form instance comes from `create_form`.
        Valid form updates the object via `update_obj`.
        If post successful, returns result of `on_add_edit_success`.
        If post failure, runs `on_add_edit_failure` and renders the form via `render_form`.
        If get, renders the form via `render_form`.
        """

        form = self.create_form(obj)
        if form is None:
            raise Exception('create_form returned None instead of a form instance')
        if meth == 'POST':
            if form.validate():
                result = self.update_obj(obj, form)
                db.session.commit()
                if result:
                    return self.on_add_edit_success(result, obj is not None)
            else:
                self.on_add_edit_failure(obj, obj is not None)

        return self.render_form(
            obj=obj,
            action=_('Edit') if obj else _('Create'),
            action_button_text=(
                _('Save Changes')
                if obj
                else _('Create {name}').format(name=self.object_name)
            ),
            form=form
        )

    def fetch_orm_obj(self, obj_id):
        return self.orm_cls.query.get(obj_id)

    def init_object(self, obj_id, action=None):
        """Load record from ORM for edit/delete cases.

        Forces 404 response if the record does not exist.

        Additional object-loading customization can be provided on action-specific hooks
        `init_object_edit` and `init_object_delete`. These methods will take no parameters, but
        they may assume `self.objinst` refers to the requested entity.
        """
        if obj_id is None:
            flask.abort(400)
        self.objinst = self.fetch_orm_obj(obj_id)
        if not self.objinst:
            flask.abort(404)
        if action:
            # Sometimes an action has particular requirements that are different from other actions.
            # For instance, a delete may be disallowed based on some property of the object itself,
            # which can't be trapped in permissions checks. Now that the object is loaded, see if
            # the view has an action init method.
            action_method = 'init_object_{}'.format(action)
            if hasattr(self, action_method):
                getattr(self, action_method)()
        return self.objinst

    def add(self):
        """View method for add. Enforce permissions and call `add_edit`."""
        return requires_permissions(self.permissions['add'])(self.add_edit)(flask.request.method)

    def edit(self, objid):
        """View method for edit. Enforce permissions, load the record, and call `add_edit`."""
        def action():
            obj = self.init_object(objid, 'edit')
            return self.add_edit(flask.request.method, obj)

        return requires_permissions(self.permissions['edit'])(action)()

    def delete(self, objid):
        """View method for delete. Enforce permissions, load the record, run ORM delete.

        If delete succeeds, return result of `on_delete_success`.
        If delete fails, return result of `on_delete_failure`.
        """
        def action():
            self.init_object(objid, 'delete')

            try:
                self.orm_cls.delete(objid)
            except sa.exc.IntegrityError:
                return self.on_delete_failure()

            return self.on_delete_success()

        return requires_permissions(self.permissions['delete'])(action)()

    def list(self):
        """View method for list. Enforce permissions, then render grid via `render_grid`."""
        return requires_permissions(self.permissions['list'])(self.render_grid)()

    @property
    def list_url_with_session(self):
        """Return list url with the session_key from the request, to support webgrid sessions."""
        return flask.url_for(self.endpoint_for_action('list'),
                             session_key=flask.request.args.get('session_key'))

    def flash_success(self, verb):
        """Add a flask flash message for success with the given `verb`."""
        # i18n: this may require reworking in order to support proper
        #       sentence structures...
        flash(_('Successfully {verb} {object}').format(
            verb=verb, object=self.object_name), 'success'
        )

    def on_delete_success(self):
        """Flash a delete success message, and redirect to list view."""
        self.flash_success(_('removed'))
        return flask.redirect(self.list_url_with_session)

    def on_delete_failure(self):
        """Flash a delete failure message, and redirect to list view."""
        flash(
            _('Unable to delete {name}. It may be referenced by other items.').format(
                name=self.object_name
            ),
            'warning'
        )
        return flask.redirect(self.list_url_with_session)

    def on_add_edit_success(self, entity, is_edit):
        """Flash an add/edit success message, and redirect to list view."""
        self.flash_success(
            _('modified')
            if is_edit
            else _('created')
        )
        return flask.redirect(self.list_url_with_session)

    def on_add_edit_failure(self, entity, is_edit):
        """Flash an add/edit message. No redirect in this case."""
        flash(_('Form errors detected.  Please see below for details.'), 'error')

    @classmethod
    def endpoint_for_action(cls, action):
        """Compute the flask endpoint for the given CRUD action."""
        return '{}:{}'.format(cls.calc_endpoint(), case_cw2dash(action))

    def make_grid(self):
        """Create an instance of `grid_cls` and initialize from request.

        Returns a grid instance."""
        grid = self.grid_cls()
        grid.apply_qs_args()
        return grid

    @property
    def grid_page_heading(self):
        """Allows customization of grid heading. Defaults to `page_title`."""
        return self.page_title(_('list'))

    def post_args_grid_setup(self, grid):
        """Apply changes to grid instance after QS args/session are loaded."""
        return grid

    def grid_template_args(self, arg_dict):
        """Allows customization of jinja template args for list view.

        `arg_dict` contains the default arguments, including anything set with `self.assign`.

        Must return a dict of template args.
        """
        return arg_dict

    def on_render_limit_exceeded(self, grid):
        """Flash a message for webgrid limit exceeded case.

        This gets run in export cases where more records are in the set than the
        file format can support."""
        flask.flash(_('Too many records to export as {}').format(grid.export_to), 'error')

    def render_grid(self):
        """Renders the grid template.

        Grid instance comes from `make_grid`.
        Grid instance may be customized via `post_args_grid_setup`.
        If grid is set to export, give that response or handle the limit exceeded error.
        Otherwise, render `grid_template` with `grid_template_args`.
        """
        grid = self.make_grid()
        grid = self.post_args_grid_setup(grid)

        if grid.session_on and flask.request.method.lower() == 'post':
            return flask.redirect(self.list_url_with_session)

        if grid.export_to:
            import webgrid

            try:
                return grid.export_as_response()
            except webgrid.renderers.RenderLimitExceeded:
                self.on_render_limit_exceeded(grid)

        title_var = flask.current_app.config.get('KEGAUTH_TEMPLATE_TITLE_VAR')

        # args added with self.assign should be passed through here
        template_args = self.grid_template_args(dict(self.template_args, **{
            'add_url': self.add_url_with_session(grid.session_key),
            title_var: self.page_title(_('list')),
            'page_heading': self.grid_page_heading,
            'object_name': self.object_name,
            'grid': grid,
        }))

        return flask.render_template(self.grid_template, **template_args)

    def add_url_with_session(self, session_key):
        """Return add url with the session_key from the request, to support webgrid sessions."""
        return flask.url_for(self.endpoint_for_action('add'), session_key=session_key)

    def cancel_url(self):
        """Return list url with the session_key from the request, to support webgrid sessions."""
        return self.list_url_with_session


class AuthRespondedView(keg.web.BaseView):
    """ Base for views which will refer out to the login authenticator for responders

        URL gets calculated from the responder class and must be a class attribute there.

        Note: if the login authenticator doesn't have the referenced key, the view will 404.
    """
    responder_key = None
    auth_manager = None
    auth_manager_key = 'login_authenticator'

    def __init__(self):
        super(AuthRespondedView, self).__init__()
        self.responding_method = 'responder'

    @classmethod
    def calc_url(cls, **kwargs):
        """Leans on login authenticator's responders to provide a URL."""
        authenticator_cls = getattr(cls.auth_manager, f'{cls.auth_manager_key}_cls')
        responder_cls = authenticator_cls.responder_cls.get(cls.responder_key)
        return getattr(responder_cls, 'url', None)

    def on_missing_responder(self):
        """Handler for requests that do not match a responder in authenticator.

        By default, aborts with 404 response."""
        flask.abort(404)

    def responder(self, *args, **kwargs):
        """Refer all requests to the responder and return the response.

        If no responder, call `on_missing_responder`."""
        authenticator = getattr(flask.current_app.auth_manager, self.auth_manager_key, None)
        if not authenticator:
            self.on_missing_responder()

        responder = authenticator.get_responder(self.responder_key)

        if not responder:
            self.on_missing_responder()

        return responder(*args, **kwargs)

    def get(self):
        # needed in keg to set up a GET route
        pass

    def post(self):
        # needed in keg to set up a POST route
        pass

    def head(self):
        # needed in keg to set up a HEAD route
        pass


class Login(AuthRespondedView):
    """Login view that uses the login authenticator's responders."""
    responder_key = 'login'


class OAuthLogin(AuthRespondedView):
    """Login view that uses the OAuth authenticator's responders."""
    responder_key = 'login'
    auth_manager_key = 'oauth_authenticator'

    @classmethod
    def calc_endpoint(cls, use_blueprint=True):
        prefix = (cls.blueprint.name + '.') if cls.blueprint and use_blueprint else ''
        return prefix + 'oauth-login'


class OAuthAuthorize(AuthRespondedView):
    """Authorization view that uses the OAuth authenticator's responders.

    Completes the OAuth login flow."""
    responder_key = 'authorize'
    auth_manager_key = 'oauth_authenticator'

    @classmethod
    def calc_endpoint(cls, use_blueprint=True):
        prefix = (cls.blueprint.name + '.') if cls.blueprint and use_blueprint else ''
        return prefix + 'oauth-authorize'


class ForgotPassword(AuthRespondedView):
    """Forgot Password view that uses the login authenticator's responders."""
    responder_key = 'forgot-password'


class ResetPassword(AuthRespondedView):
    """Reset Password view that uses the login authenticator's responders."""
    responder_key = 'reset-password'


class VerifyAccount(AuthRespondedView):
    """Verification view that uses the login authenticator's responders."""
    responder_key = 'verify-account'


class Logout(AuthRespondedView):
    """Logout view that uses the login authenticator's responders."""
    responder_key = 'logout'

    # Do this manually here, rather than using requires_user, because we want
    #   better control over what happens if an unauthenticated user uses the
    #   view. But, this prevents the route from being present in nav if a user
    #   is not logged in.
    __keg_auth_requires_user__ = True


@requires_permissions('auth-manage')
class User(CrudView):
    """Default User CRUD view. Uses auth-manage permission for all targets."""
    url = '/users'
    object_name = _('User')
    object_name_plural = _('Users')
    form_cls = staticmethod(forms.user_form)

    def create_form(self, obj):
        form_cls = self.form_cls(flask.current_app.config,
                                 allow_superuser=flask_login.current_user.is_superuser,
                                 endpoint=self.endpoint_for_action('edit'))
        return form_cls(obj=obj)

    @keg.web.route(post_only=True)
    def resend_verification_email(self):
        validate_csrf(flask.request.form['csrf_token'])
        auth_manager = keg.current_app.auth_manager
        if not auth_manager.login_authenticator.is_domain_excluded(flask.request.form['user_id']):
            auth_manager.resend_verification_email(flask.request.form['user_id'])
            flask.flash(str(_('Verification email has been sent')), 'success')
        return flask.redirect(flask.url_for(self.endpoint_for_action('list')))

    @property
    def orm_cls(self):
        return flask.current_app.auth_manager.entity_registry.user_cls

    @property
    def grid_cls(self):
        return grids.make_user_grid(
            edit_endpoint=self.endpoint_for_action('edit'),
            edit_permission=self.permissions['edit'],
            delete_endpoint=self.endpoint_for_action('delete'),
            delete_permission=self.permissions['delete'],
            resend_verification_endpoint=self.endpoint_for_action('resend-verification-email')
        )

    def create_user(self, form):
        auth_manager = keg.current_app.auth_manager
        email_enabled = (
            flask.current_app.config.get('KEGAUTH_EMAIL_OPS_ENABLED', True)
            and hasattr(form, 'send_welcome')
            and form.send_welcome.data
        )
        user_kwargs = {}
        user_kwargs['mail_enabled'] = email_enabled
        for field in form.data:
            # Only want fields that are on the class in kwargs
            # if we pass other stuff like permission_ids
            # user model wont be saved
            if hasattr(self.orm_cls, field):
                user_kwargs[field] = form[field].data
        user_kwargs['permissions'] = form.get_selected_permissions()
        user_kwargs['bundles'] = form.get_selected_bundles()
        user_kwargs['groups'] = form.get_selected_groups()
        obj = auth_manager.create_user(user_kwargs, _commit=False)
        return obj

    def update_obj(self, obj, form):
        if obj is None:
            obj = self.create_user(form)
        else:
            form.populate_obj(obj)
            obj.permissions = form.get_selected_permissions()
            obj.bundles = form.get_selected_bundles()
            obj.groups = form.get_selected_groups()
        # only reset a password if it is on the form and populated
        if hasattr(form, 'reset_password') and form.reset_password.data:
            obj.password = form.reset_password.data

        return obj

    def delete(self, objid):
        # ensure user cannot delete oneself
        if objid == flask_login.current_user.id:
            return self.on_delete_failure()
        return super(User, self).delete(objid)


@requires_permissions('auth-manage')
class Group(CrudView):
    """Default Group CRUD view. Uses auth-manage permission for all targets."""
    url = '/groups'
    object_name = _('Group')
    object_name_plural = _('Groups')
    form_cls = staticmethod(forms.group_form)

    def create_form(self, obj):
        form_cls = self.form_cls(endpoint=self.endpoint_for_action('edit'))
        return form_cls(obj=obj)

    @property
    def orm_cls(self):
        return flask.current_app.auth_manager.entity_registry.group_cls

    @property
    def grid_cls(self):
        return grids.make_group_grid(
            edit_endpoint=self.endpoint_for_action('edit'),
            edit_permission=self.permissions['edit'],
            delete_endpoint=self.endpoint_for_action('delete'),
            delete_permission=self.permissions['delete']
        )

    def update_obj(self, obj, form):
        obj = obj or self.add_orm_obj()
        form.populate_obj(obj)
        obj.permissions = form.get_selected_permissions()
        obj.bundles = form.get_selected_bundles()
        return obj


@requires_permissions('auth-manage')
class Bundle(CrudView):
    """Default Bundle CRUD view. Uses auth-manage permission for all targets."""
    url = '/bundles'
    object_name = _('Bundle')
    object_name_plural = _('Bundles')
    form_cls = staticmethod(forms.bundle_form)

    def create_form(self, obj):
        form_cls = self.form_cls(endpoint=self.endpoint_for_action('edit'))
        return form_cls(obj=obj)

    @property
    def orm_cls(self):
        return flask.current_app.auth_manager.entity_registry.bundle_cls

    @property
    def grid_cls(self):
        return grids.make_bundle_grid(
            edit_endpoint=self.endpoint_for_action('edit'),
            edit_permission=self.permissions['edit'],
            delete_endpoint=self.endpoint_for_action('delete'),
            delete_permission=self.permissions['delete']
        )

    def update_obj(self, obj, form):
        obj = obj or self.add_orm_obj()
        form.populate_obj(obj)
        obj.permissions = form.get_selected_permissions()
        return obj


@requires_permissions('auth-manage')
class Permission(keg.web.BaseView):
    """Default Permission view. Uses auth-manage permission."""
    url = '/permissions'
    grid_template = 'keg-auth/crud-list.html'

    @property
    def grid_cls(self):
        return grids.make_permission_grid()

    def get(self):
        grid = self.grid_cls()
        grid.apply_qs_args()

        if grid.export_to:
            return grid.export_as_response()

        title_var = flask.current_app.config.get('KEGAUTH_TEMPLATE_TITLE_VAR')
        self.assign(title_var, _('Permissions'))

        return flask.render_template(
            self.grid_template,
            page_heading=_('Permissions'),
            grid=grid,
            **self.template_args,
        )


def make_blueprint(import_name, _auth_manager, bp_name='auth', login_cls=Login,
                   forgot_cls=ForgotPassword, reset_cls=ResetPassword, logout_cls=Logout,
                   verify_cls=VerifyAccount, user_crud_cls=User, group_crud_cls=Group,
                   bundle_crud_cls=Bundle, permission_cls=Permission, oauth_login_cls=OAuthLogin,
                   oauth_auth_cls=OAuthAuthorize, blueprint_class=flask.Blueprint, **kwargs):
    """ Blueprint factory for keg-auth views

        Most params are assumed to be view classes. `_auth_manager` is the extension instance meant
        for the app on which this blueprint will be used: it is necessary in order to apply url
        routes for user functions.

        blueprint_class is the class to be instantiated as the Flask blueprint for auth views. The
        default is flask.blueprint, but a custom blueprint may be provided.
    """
    _blueprint = blueprint_class(bp_name, import_name, **kwargs)

    # auth responded views get the auth manager assigned for URL calculation, etc.
    # This has to happen before assign_blueprint.
    auth_responded_views = (
        login_cls, forgot_cls, reset_cls, verify_cls, logout_cls, oauth_login_cls, oauth_auth_cls
    )
    for view_cls in auth_responded_views:
        if view_cls:
            view_cls.auth_manager = _auth_manager

    for view_cls in (user_crud_cls, group_crud_cls, bundle_crud_cls, permission_cls,
                     *auth_responded_views):
        if view_cls:
            view_cls.assign_blueprint(_blueprint)

    return _blueprint
